from tkinter import Tk, StringVar, Label, Entry, Button

# Perhaps add subplots, each of a specific domain with specified
# min and max x and y values.
# Best to not allow overplotting of different quantities (v and i, etc.)
# Maybe specify domain and quantity (time, voltage) then specify
# components (or perhaps nodes)
# Could have a fixed number for the maximum number of plots, say 8
# and then have 8 dropdown lists.   This is probably easier than having a
# menu item for adding another plot.
# But also need a way to handle multiple subplots (these are not overdrawn)
# Could have a specified number of rows and cols for the subplot grid.

# Automatically add labels, R1 voltage, C1 current, etc.
# Perhaps allow expressions, R1.V(t) - R2.V(t)?

# What about real part, magnitude, phase etc?
# Could use expression such as R1.V(f).phase

# Expression  row  col  xmin  xmax  ymin  ymax colour style


class MultiplotDialog:

    def __init__(self, ui):

        self.ui = ui

        self.window = Tk()

        circuit = ui.model.circuit

        # TODO: need undefined_symbols for old Lcapy
        symbols = circuit.undefined_symbols
        if symbols != []:
            ui.show_error_dialog(
                'Undefined symbols: %s.  Use edit values to define' % ', '.join(symbols))

        elements = circuit.elements

        entries = []
        for elt in elements:
            # TODO
            print(elt)

        row = 0

        # R1  voltage  domain

        button = Button(self.window, text="OK", command=self.on_update)
        button.grid(row=row)

    def on_update(self):

        self.window.destroy()



#!/usr/bin/env python3
"""lcapy-tk V0.0.3
Copyright (c) 2023 Michael P. Hayes, UC ECE, NZ

Usage: lcapy-tk [infile.sch]
"""

from argparse import ArgumentParser
import sys
from lcapygui import LcapyTk
from lcapy import expr as lcapify


def schtex_exception(type, value, tb):
    if hasattr(sys, 'ps1') or not sys.stderr.isatty():
        # We are not in interactive mode or we don't have a tty-like
        # device, so call the default hook
        sys.__excepthook__(type, value, tb)
    else:
        import traceback
        import pdb
        # We are in interactive mode, print the exception...
        traceback.print_exception(type, value, tb)
        print()
        # ...then start the debugger in post-mortem mode.
        pdb.pm()


def main(argv=None):

    if argv is None:
        argv = sys.argv

    parser = ArgumentParser(
        description='Generate lcapy netlists.')
    parser.add_argument('--version', action='version',
                        version=__doc__.split('\n')[0])
    parser.add_argument('--pdb', action='store_true',
                        default=False,
                        help="enter python debugger on exception")
    parser.add_argument('--debug', type=int,
                        dest='debug', default=None,
                        help="enable debugging")
    parser.add_argument('--expr', type=str,
                        dest='expr', default=None,
                        help="Lcapy expression")
    parser.add_argument('filenames', type=str, nargs='*',
                        help='schematic filename(s)', default=[])

    args = parser.parse_args()

    if args.pdb:
        sys.excepthook = schtex_exception

    e = LcapyTk(args.filenames, debug=args.debug)

    if args.expr is not None:
        e.show_expr_dialog(lcapify(args.expr))

    e.display()

    return 0


if __name__ == '__main__':
    sys.exit(main())



from .bipole import BipoleComponent


class Wire(BipoleComponent):

    type = 'W'
    args = ()
    has_value = False

    def draw(self, editor, sketcher, **kwargs):

        x1, y1 = self.node1.x, self.node1.y
        x2, y2 = self.node2.x, self.node2.y

        kwargs = self.make_kwargs(editor, **kwargs)
        sketcher.stroke_line(x1, y1, x2, y2, **kwargs)



from .bipole import BipoleComponent


class Diode(BipoleComponent):

    type = 'D'
    args = ()
    kinds = {'-': '', '-led': 'LED', '-photo': 'Photo', '-schottky': 'Schottky',
             '-zener': 'Zener', '-zzener': 'Zzener', '-tunnel': 'Tunnel',
             '-varcap': 'VarCap', '-bidirectional': 'Bidirectional',
             '-tvs': 'TVS', '-laser': 'Laser'}
    styles = {'empty': 'Empty', 'full': 'Full', 'stroke': 'Stroke'}
    default_kind = '-'
    default_style = 'empty'
    has_value = False



from xml.dom import minidom
from svgpathtools.parser import parse_transform
from svgpath2mpl import parse_path
from numpy import all

# TODO: look for style attribute, split by semicolon.  One of the most
# useful is fill, for example, 'fill:rgb(0%,0%,0%)' or 'fill:none'.
# The paths and use nodes have a style attribute.

# The SVG file has:
# 1. Symbol definitions
# 2. Clip paths
# 3. Drawing paths
# 4. Symbol uses
#
# It uses the <g> element to group other elements


def parse_style(string):

    style = {}
    parts = string.strip().split(';')
    for part in parts:
        bits = part.split(':')
        if len(bits) == 2:
            k, v = bits
            style[k] = v
    return style


class SVGPath:

    def __init__(self, path, transform, style, symbol):

        self.path = path
        self.transform = transform
        self.symbol = symbol
        self.style = style


class SVGParse:

    def __init__(self, filename):

        self.filename = filename

        doc = minidom.parse(filename)

        svg_paths = doc.getElementsByTagName('path')
        svg_defs = doc.getElementsByTagName('defs')
        svg_uses = doc.getElementsByTagName('use')

        svg = doc.getElementsByTagName('svg')[0]
        width_str = svg.getAttribute('width')
        height_str = svg.getAttribute('height')
        if not width_str.endswith('pt'):
            raise ValueError('Need to to handle other units.')
        if not height_str.endswith('pt'):
            raise ValueError('Need to to handle other units.')

        self.width = float(width_str[:-2])
        self.height = float(height_str[:-2])

        # Ignore paths for symbol defs and clip paths
        svg_paths = [path for path in svg_paths
                     if path.parentNode.tagName not in ('symbol', 'clipPath')]

        svg_ds = [path.getAttribute('d') for path in svg_paths]
        svg_styles = [path.getAttribute('style') for path in svg_paths]
        svg_transforms = [path.getAttribute('transform') for path in svg_paths]

        transforms = []
        for transform in svg_transforms:
            if transform == '':
                transform = 'matrix(1,0,0,1,0,0)'
            transforms.append(transform)
        svg_transforms = transforms

        self.paths = []
        for d, transform, style in zip(svg_ds, svg_transforms, svg_styles):
            path = parse_path(d)
            transform = parse_transform(transform)
            self.paths.append(
                SVGPath(path, transform, parse_style(style), False))

        if svg_defs != []:
            svg_symbols = svg_defs[0].getElementsByTagName('symbol')
            symbols = {}
            for symbol in svg_symbols:
                symbol_id = symbol.getAttribute('id')
                path = symbol.getElementsByTagName('path')[0]
                d = path.getAttribute('d')
                symbols[symbol_id] = d

            for use in svg_uses:
                symbol_id = use.getAttribute('xlink:href')[1:]
                x = use.getAttribute('x')
                y = use.getAttribute('y')
                transform = 'matrix(1,0,0,1,%s,%s)' % (x, y)
                path = parse_path(symbols[symbol_id])
                transform = parse_transform(transform)
                self.paths.append(
                    SVGPath(path, transform, parse_style(style), True))



from tkinter import Tk, Entry, Button, Text, BOTH, END
from lcapy import expr


class EditDialog:

    def __init__(self, expr, ui):

        self.expr = expr
        self.ui = ui

        self.window = Tk()
        self.window.title('Expression editor')

        s = str(expr)

        self.text = Text(self.window)
        self.text.pack(fill=BOTH, expand=1)
        self.text.insert(END, s)

        button = Button(self.window, text='Show', command=self.on_show)
        button.pack()

    def on_show(self):

        expr_str = self.text.get('1.0', END).strip()

        try:
            self.ui.show_expr_dialog(expr(expr_str))
            self.window.destroy()

        except Exception as e:
            self.ui.show_error_dialog('Cannot evaluate expression')



from pathlib import Path
import json


# Perhaps make a dict?
class Preferences:

    def __init__(self):

        self.version = 1
        self.label_nodes = 'none'
        self.draw_nodes = 'connections'
        self.label_cpts = 'name'
        self.style = 'american'
        self.node_size = 0.1
        self.node_color = 'black'
        self.grid = 'on'
        self.lw = 1.2
        self.show_units = 'false'
        self.xsize = 36
        self.ysize = 22
        self.snap_grid = 'true'
        self.voltage_dir = 'RP'

        self.load()

    @property
    def _dirname(self):

        return Path('~/.lcapy/').expanduser()

    @property
    def _filename(self):

        return self._dirname / 'preferences.json'

    def load(self):

        dirname = self._dirname
        if not dirname.exists():
            return

        s = self._filename.read_text()
        d = json.loads(s)
        for k, v in d.items():
            setattr(self, k, v)

    def save(self):

        dirname = self._dirname
        if not dirname.exists():
            dirname.mkdir()
        s = json.dumps(self, default=lambda o: o.__dict__,
                       sort_keys=True, indent=4)

        self._filename.write_text(s)

    def schematic_preferences(self):

        opts = ('draw_nodes', 'label_nodes', 'style', 'voltage_dir')

        foo = []
        for opt in opts:
            foo.append(opt + '=' + getattr(self, opt))
        s = ', '.join(foo)

        if self.label_cpts == 'name':
            s += ', label_ids=true'
            s += ', label_values=false'
        elif self.label_cpts == 'value':
            s += ', label_ids=false'
            s += ', label_values=true'
        elif self.label_cpts == 'value+name':
            s += ', label_ids=true'
            s += ', label_values=true'

        return s



from tkinter import Tk, Text, END


class MessageDialog:

    def __init__(self, message, title=''):

        self.window = Tk()
        self.window.title(title)

        text = Text(self.window)
        text.pack()

        text.insert(END, message)



from .menu import MenuBar, MenuDropdown, MenuItem
from .expr_calc import ExprCalc
from .exprimage import ExprImage
from lcapy import Expr, ExprTuple, Matrix
from PIL import Image, ImageTk
from tkinter import Tk, Button, Label, Frame, BOTH, X


def make_python(expr):
    """Return Python code that can be evaluated to create the
        expression."""

    # Output imports?

    symbols = expr.symbols

    s = ''
    for symbol in symbols:
        # Skip domain variables
        if symbol in ('f', 's', 't', 'w', 'omega',
                      'jf', 'jw', 'jomega', 'n', 'k', 'z'):
            continue

        # TODO, add other assumptions
        if symbols[symbol].is_positive:
            s += "%s = symbol('%s', positive=True)\n" % (symbol, symbol)
        else:
            s += "%s = symbol('%s')\n" % (symbol, symbol)

    s += repr(expr)
    return s


class ExprDialog(Tk):

    def __init__(self, expr, ui, title=''):

        super().__init__()

        self.report_callback_exception = ui.report_callback_exception
        self.expr = expr
        self.ui = ui
        self.title(title)

        mdd = None
        if isinstance(expr, Matrix):
            nrows = expr.rows
            ncols = expr.cols
            delim = '' if nrows < 10 else ','
            items = []
            for i in range(nrows):
                for j in range(ncols):
                    items.append(MenuItem('A%d%s%d' % (i + 1, delim, j + 1),
                                          self.on_element))

            mdd = MenuDropdown('Element', 0, items)

        menudropdowns = [
            MenuDropdown('File', 0,
                         [MenuItem('Save Python', self.on_save_python),
                          MenuItem('Save LaTeX', self.on_save_latex)
                          ]),
            MenuDropdown('Edit', 0,
                         [MenuItem('Expression', self.on_edit),
                          MenuItem('Python', self.on_edit_python)
                          ]),
            MenuDropdown('View', 0,
                         [MenuItem('Plot', self.on_plot),
                          MenuItem('LaTeX', self.on_latex),
                          MenuItem('Python', self.on_python),
                          MenuItem('Attributes', self.on_attributes)]),
            MenuDropdown('Manipulate', 0,
                         [MenuItem('Approximate', self.on_manipulate),
                          MenuItem('Evaluate', self.on_manipulate),
                          MenuItem('Limit', self.on_manipulate),
                          MenuItem('Parameterize', self.on_manipulate),
                          MenuItem('Poles', self.on_manipulate),
                          MenuDropdown('Simplify', 0,
                                       [MenuItem('Simplify', self.on_manipulate),
                                        MenuItem('Simplify conjugates',
                                                 self.on_manipulate),
                                        MenuItem('Simplify factors',
                                                 self.on_manipulate),
                                        MenuItem('Simplify terms',
                                                 self.on_manipulate),
                                        MenuItem('Simplify sin/cos',
                                                 self.on_manipulate),
                                        MenuItem('Simplify Dirac delta',
                                                 self.on_manipulate),
                                        MenuItem('Simplify Heaviside',
                                                 self.on_manipulate),
                                        MenuItem('Simplify rect',
                                                 self.on_manipulate),
                                        MenuItem('Rationalize denominator',
                                                 self.on_manipulate),
                                        MenuItem('Cancel terms',
                                                 self.on_manipulate),
                                        ]),
                          MenuItem('Solve', self.on_manipulate),
                          MenuItem('Substitute', self.on_manipulate),
                          MenuItem('Zeros', self.on_manipulate),
                          ]),
            MenuDropdown('Transform', 0,
                         [MenuItem('Time', self.on_transform),
                          MenuItem('Laplace', self.on_transform),
                          MenuItem('Phasor', self.on_transform),
                          MenuItem('Fourier', self.on_transform),
                          MenuItem('Angular Fourier', self.on_transform),
                          MenuItem('Frequency', self.on_transform),
                          MenuItem('Angular frequency', self.on_transform)]),
            MenuDropdown('Select', 0,
                         [mdd,
                          MenuItem('Real', self.on_select),
                          MenuItem('Imaginary', self.on_select),
                          MenuItem('Magnitude', self.on_select),
                          MenuItem('dB', self.on_select),
                          MenuItem('Sign', self.on_select),
                          MenuItem('Phase degrees', self.on_select),
                          MenuItem('Phase radians', self.on_select),
                          ]),
            MenuDropdown('Format', 0,
                         [MenuItem('ZPK', self.on_format),
                          MenuItem('Canonical', self.on_format),
                          MenuItem('Time constant', self.on_format),
                          MenuItem('Time constant terms', self.on_format),
                          MenuItem('General', self.on_format),
                          MenuItem('Standard', self.on_format),
                          MenuItem('Partial fraction', self.on_format),
                          MenuItem('Polar', self.on_format),
                          MenuItem('Cartesian', self.on_format),
                          ]),
        ]

        self.menubar = MenuBar(menudropdowns)
        self.menubar.make(self)

        png_filename = ExprImage(expr).image()
        image = Image.open(png_filename)

        self.expr_label = Label(self, text='', width=image.width + 100,
                                height=image.height + 100)
        self.expr_label.pack(fill=BOTH, expand=True)

        self.minsize(550, 100)

        self.update()

    def update(self):

        try:
            self.show_img(self.expr)
        except Exception as e:
            self.expr_label.config(text=e)

    def show_img(self, e, pad=30):

        # TODO, fixme
        # if self.ui.model.preferences.show_units == 'true':
        #    e = e * e.units

        png_filename = ExprImage(e).image()

        image = Image.open(png_filename)

        right = pad
        left = pad
        top = pad
        bottom = pad

        width, height = image.size

        new_width = width + right + left
        new_height = height + top + bottom

        # Add border
        background = (245, 245, 245)
        image_pad = Image.new(image.mode, (new_width, new_height), background)
        image_pad.paste(image, (left, top))

        img_pad = ImageTk.PhotoImage(image_pad, master=self)

        self.expr_label.config(image=img_pad)
        self.expr_label.photo = img_pad

    def apply_attribute(self, attributes, arg):

        try:
            attribute = attributes[arg]

            e = ExprCalc(self.expr)
            expr = e.attribute(attribute)
            self.ui.show_expr_dialog(expr, title=self.title)
        except Exception as e:
            self.ui.show_error_dialog(e)

    def apply_method(self, methods, arg):

        try:
            method = methods[arg]

            e = ExprCalc(self.expr)
            expr = e.method(method)
            self.ui.show_expr_dialog(expr, title=self.title)
        except Exception as e:
            self.ui.show_error_dialog(e)

    def on_attributes(self, arg):

        self.ui.show_expr_attributes_dialog(self.expr, title=self.title)

    def on_edit(self, arg):

        self.ui.show_edit_dialog(self.expr)

    def on_edit_python(self, arg):

        self.ui.show_python_dialog(self.expr)

    def on_format(self, arg):

        formats = {'Canonical': 'canonical',
                   'Standard': 'standard',
                   'General': 'general',
                   'Time constant': 'timeconst',
                   'Time constant terms': 'timeconst_terms',
                   'ZPK': 'ZPK',
                   'Partial fraction': 'partfrac',
                   'Time constant': 'timeconst',
                   'Polar': 'polar',
                   'Cartesian': 'cartesian'}

        if arg in ('Polar', 'Cartesian'):
            # What was I thinking?
            self.apply_attribute(formats, arg)
        else:
            self.apply_method(formats, arg)

    def on_latex(self, arg):

        self.ui.show_message_dialog(self.expr.latex())

    def on_manipulate(self, arg):

        simplify = {'Simplify': 'simplify',
                    'Simplify conjugates': 'simplify_conjugates',
                    'Simplify factors': 'simplify_factors',
                    'Simplify terms': 'simplify_terms',
                    'Simplify sin/cos': 'simplify_sin_cos',
                    'Simplify Dirac delta': 'simplify_dirac_delta',
                    'Simplify Heaviside': 'simplify_heaviside',
                    'Simplify rect': 'simplify_rect',
                    'Cancel terms': 'cancel_terms',
                    'Rationalize denominator': 'rationalize_denominator'}

        try:
            if arg == 'Approximate':
                self.ui.show_approximate_dialog(self.expr, title=self.title)
            elif arg == 'Evaluate':
                self.ui.show_expr_dialog(
                    self.expr.evaluate(), title=self.title)
            elif arg == 'Limit':
                self.ui.show_limit_dialog(self.expr, title=self.title)
            elif arg == 'Parameterize':
                self.ui.show_expr_dialog(ExprTuple(self.expr.parameterize()),
                                         title=self.title)
            elif arg == 'Poles':
                self.ui.show_expr_dialog(
                    self.expr.poles(), title=self.title)
            elif arg in simplify:
                self.apply_method(simplify, arg)
            elif arg == 'Solve':
                self.ui.show_expr_dialog(
                    self.expr.solve(), title=self.title)
            elif arg == 'Substitute':
                self.ui.show_subs_dialog(self.expr, title=self.title)
            elif arg == 'Zeros':
                self.ui.show_expr_dialog(
                    self.expr.zeros(), title=self.title)
        except Exception as e:
            self.ui.show_error_dialog(e)

    def on_save_latex(self, arg):

        pathname = self.ui.save_file_dialog('expr.tex', doc='LaTeX file',
                                            ext='*.tex')
        if pathname == '' or pathname == ():
            return

        with open(pathname, 'w') as f:
            f.write(self.expr.latex())

    def on_save_python(self, arg):

        pathname = self.ui.save_file_dialog('expr.py', doc='Python file',
                                            ext='*.py')
        if pathname == '' or pathname == ():
            return

        with open(pathname, 'w') as f:
            f.write(make_python(self.expr))

    def on_select(self, arg):

        parts = {'Real': 'real',
                 'Imaginary': 'imag',
                 'Magnitude': 'magnitude',
                 'dB': 'dB',
                 'Sign': 'sign',
                 'Phase degrees': 'phase_degrees',
                 'Phase radian': 'phase_radians'}

        self.apply_attribute(parts, arg)

    def on_element(self, arg):

        arg = arg[1:]

        if ',' in arg:
            r, c = arg.split(',')
        else:
            r = arg[0]
            c = arg[1]

        row = int(r) - 1
        col = int(c) - 1

        e = ExprCalc(self.expr)
        expr = e.eval('[%s, %d]' % (row, col))
        self.ui.show_expr_dialog(expr, title=self.title)

    def on_plot(self, arg):

        if not isinstance(self.expr, Expr):
            self.ui.info_dialog('Cannot plot expression')
            return

        self.ui.show_plot_properties_dialog(self.expr)

    def on_python(self, arg):

        s = make_python(self.expr)

        self.ui.show_message_dialog(s, 'Python expression')

    def on_transform(self, arg):

        domains = {'Time': 'time',
                   'Phasor': 'phasor',
                   'Laplace': 'laplace',
                   'Fourier': 'fourier',
                   'Frequency': 'frequency_response',
                   'Angular Fourier': 'angular_fourier',
                   'Angular frequency': 'angular_frequency_response'}

        self.apply_method(domains, arg)



from tkinter import Tk
from tkinter.ttk import Button, Label
from PIL import Image, ImageTk

from .lateximage import LatexImage


# See https://stackoverflow.com/questions/56043767/
# show-large-image-using-scrollbar-in-python

class EquationsDialog:

    def __init__(self, expr, ui, title=''):

        self.expr = expr
        self.ui = ui
        self.labelentries = None
        self.title = title

        s = '\\begin{tabular}{ll}\n'

        for k, v in expr.items():
            if not isinstance(k, str):
                k = k.latex()

            s += '$' + k + '$: & $' + v.latex() + '$\\\\ \n'

        s += '\\end{tabular}\n'
        self.s = s

        self.window = Tk()
        self.window.title(title)

        self.expr_label = Label(self.window, text='')
        self.expr_label.grid(row=0)

        button = Button(self.window, text="LaTeX", command=self.on_latex)
        button.grid(row=1, sticky='w')

        self.update()

    def update(self):

        try:
            self.show_img()
        except Exception as e:
            self.expr_label.config(text=e)

    def show_img(self):

        png_filename = LatexImage(self.s).image()
        img = ImageTk.PhotoImage(Image.open(png_filename), master=self.window)
        self.expr_label.config(image=img)
        self.expr_label.photo = img

    def on_latex(self):

        self.ui.show_message_dialog(self.s)



from .transistor import Transistor


class BJT(Transistor):

    type = "Q"
    angle_offset = 90
    default_kind = 'npn-'

    kinds = {'npn-': 'NPN',
             'pnp-': 'PNP',
             'npn-nigbt': 'NPN IGBT',
             'pnp-pigbt': 'PNP IGBT gate',
             'npn-bodydiode': 'NPN with body diode',
             'pnp-bodydiode': 'PNP with body diode',
             'npn-nigbt-bodydiode': 'NPN IGBT with body diode',
             'pnp-pigbt-bodydiode': 'PNP IGBT with body diode',
             'npn-Lnigbt': 'L-shaped NPN IGBT',
             'pnp-Lpigbt':  'L-shaped PNP IGBT'}

    # TODO: add base offset for Lnigbt, Lpigbt



"""
Defines the components that lcapy-gui can draw
"""

from numpy import array, dot
from numpy.linalg import norm
from lcapy.opts import Opts

from typing import Union
from abc import ABC, abstractmethod
from math import sqrt, degrees, atan2


class Component(ABC):

    """
    Describes an lcapy-gui component.
    This is an abstract class, specific components are derived from here.
    """

    args = ('Value', )
    kinds = {}
    styles = {}
    can_stretch = False
    default_kind = ''
    default_style = ''
    label_offset = 0.6
    angle_offset = 0
    fields = {'label': 'Label',
              'voltage_label': 'Voltage label',
              'current_label': 'Current label',
              'flow_label': 'Flow label',
              'color': 'Color',
              'attrs': 'Attributes'}
    extra_fields = {}
    has_value = True

    voltage_keys = ('v', 'v_', 'v^', 'v_>', 'v_<', 'v^>', 'v^<',
                    'v<', 'v>')
    current_keys = ('i', 'i_', 'i^', 'i_>',  'i_<', 'i^>', 'i^<',
                    'i>_', 'i<_', 'i>^', 'i<^', 'i>', 'i<', 'ir')
    flow_keys = ('f', 'f_', 'f^', 'f_>',  'f_<', 'f^>', 'f^<',
                 'f>_', 'f<_', 'f>^', 'f<^', 'f>', 'f<')
    label_keys = ('l', 'l_', 'l^')
    annotation_keys = ('a', 'a_', 'a^')

    # These are not passed to the sktecher.
    ignore_keys = ('left', 'right', 'up', 'down', 'size', 'rotate',
                   'pinnodes', 'pinnames', 'pins', 'pinlabels',
                   'mirrorinputs', 'free', 'ignore', 'nosim', 'arrow',
                   'startarrow', 'endarrow', 'bus', 'anchor', 'fixed')

    all_keys = voltage_keys + current_keys + flow_keys + \
        label_keys + annotation_keys + ignore_keys

    # TODO: add class methods to construct Component from
    # an Lcapy cpt or from a cpt type.

    def __init__(self, kind='', style='', name=None, nodes=None, opts=None):

        if nodes is None:
            nodes = []
        if opts is None:
            opts = Opts()
        else:
            opts = opts.copy()

        self.name = name
        self.nodes = nodes
        self.opts = opts
        self.control = None
        self.attrs = ''
        self.annotations = []
        self.label = ''
        self.voltage_label = ''
        self.current_label = ''
        self.flow_label = ''
        self.color = ''
        self.scale = 1

        self.mirror = False
        self.invert = False

        if kind == '':
            kind = self.default_kind
        self.kind = kind
        self.inv_kinds = {v: k for k, v in self.kinds.items()}

        if style == '':
            style = self.default_style
        self.style = style
        self.inv_styles = {v: k for k, v in self.styles.items()}

        for k, v in self.extra_fields.items():
            if k in opts:
                opts.remove(k)
                setattr(self, k, True)

        opts = self.filter_opts(opts)

        parts = []
        for k, v in opts.items():
            if k in ('color', 'colour'):
                self.color = v
            elif k == 'scale':
                try:
                    self.scale = float(v)
                except:
                    self.scale = 1.0
            elif k == 'mirror':
                self.mirror = True
            elif k == 'invert':
                self.invert = True
            elif k == 'kind':
                self.kind += '-' + v
            elif k == 'style':
                self.style = v
            elif k in self.all_keys:
                pass
            else:
                if v == '':
                    parts.append(k)
                else:
                    parts.append(k + '=' + v)
        self.attrs = ', '.join(parts)

    def filter_opts(self, opts):

        connection_keys = ('input', 'output', 'bidir', 'pad')
        ground_keys = ('ground', 'sground', 'rground',
                       'cground', 'nground', 'pground', '0V')
        supply_positive_keys = ('vcc', 'vdd')
        supply_negative_keys = ('vee', 'vss')
        io_keys = ('input', 'output', 'bidir')
        supply_keys = supply_positive_keys + supply_negative_keys
        implicit_keys = ('implicit', ) + ground_keys + supply_keys + io_keys

        stripped = list(opts.strip(*implicit_keys))
        if len(stripped) > 1:
            raise ValueError('Multiple connection kinds: ' +
                             ', '.join(stripped))
        elif len(stripped) == 1:
            kind = stripped[0]
            if kind == 'implicit':
                kind = 'ground'
            self.kind = '-' + kind

        return opts

    @property
    @classmethod
    @abstractmethod
    def type(cls) -> str:
        """
        Component type identifer used by lcapy.
        E.g. Resistors have the identifier R.
        """
        ...

    def __str__(self) -> str:

        return self.type + ' ' + '(%s, %s) (%s, %s)' % \
            (self.node1.pos.x, self.node1.pos.y,
             self.node2.pos.x, self.node2.pos.y)

    @property
    def sketch_key(self):

        s = self.type
        s += '-' + self.cpt_kind
        s += '-' + self.symbol_kind
        s += '-' + self.style
        s = s.strip('-')
        return s

    @property
    def cpt_kind(self):

        parts = self.kind.split('-')
        return parts[0]

    @property
    def symbol_kind(self):

        parts = self.kind.split('-')
        return '-'.join(parts[1:])

    @property
    def labelled_nodes(self):

        return self.nodes

    @property
    def drawn_nodes(self):

        return self.nodes

    def draw(self, editor, sketcher, **kwargs):
        """
        Handles drawing specific features of components.
        """

        # Handle ports where nothing is drawn.
        if self.sketch is None:
            return

        x1, y1 = self.node1.x, self.node1.y
        x2, y2 = self.node2.x, self.node2.y
        dx = x2 - x1
        dy = y2 - y1

        r = self.length
        if r == 0:
            editor.ui.show_warning_dialog(
                'Ignoring zero size component ' + self.name)
            return

        angle = self.angle

        # Width in cm
        w = self.sketch.width / 72 * 2.54

        p1 = array((x1, y1))
        if r != 0:
            dw = array((dx, dy)) / r * (r - w) / 2
            p1p = p1 + dw
        else:
            # For zero length wires
            p1p = p1

        kwargs = self.make_kwargs(editor, **kwargs)

        if 'invisible' in kwargs or 'nodraw' in kwargs or 'ignore' in kwargs:
            return

        sketcher.sketch(self.sketch, offset=p1p, angle=angle,
                        snap=True, **kwargs)

        # Add stretchable wires
        if self.can_stretch:

            p2 = array((x2, y2))
            p2p = p2 - dw

            # TODO: generalize
            kwargs.pop('mirror', False)
            kwargs.pop('invert', False)

            sketcher.stroke_line(*p1, *p1p, **kwargs)
            sketcher.stroke_line(*p2p, *p2, **kwargs)

            # For transistors.
            if len(self.nodes) == 3:

                x3, y3 = self.nodes[1].x, self.nodes[1].y

                mx = (x1 + x2) / 2
                my = (y1 + y2) / 2
                dx = mx - x3
                dy = my - y3
                r = sqrt(dx**2 + dy**2)

                p3 = array((x3, y3))

                # Height in cm
                h = self.sketch.height / 72 * 2.54
                dh = array((dx, dy)) / r * (r - h)
                p3p = p3 + dh
                sketcher.stroke_line(*p3, *p3p, **kwargs)

        # TODO, add label, voltage_label, current_label, flow_label

    def make_kwargs(self, editor, **kwargs):

        opts = Opts(self.attrs)

        kwargs['lw'] = kwargs.pop('lw', editor.preferences.lw)

        for k, v in opts.items():
            if k in ('bodydiode', ):
                continue
            if v == '':
                v = True
            kwargs[k] = v

        if kwargs.pop('thick', False):
            kwargs['lw'] = kwargs['lw'] * 2

        if self.color != '':
            kwargs['color'] = self.color

        if self.mirror:
            kwargs['mirror'] = True

        if self.invert:
            kwargs['invert'] = True

        if kwargs.pop('dashed', False):
            kwargs['linestyle'] = '--'

        if kwargs.pop('dotted', False):
            kwargs['linestyle'] = ':'

        return kwargs

    @property
    def length(self) -> float:
        """
        Computes the length of the component.
        """
        return (self.node2.pos - self.node1.pos).norm()

    @property
    def midpoint(self):
        """
        Computes the midpoint of the component.
        """

        return (self.node1.pos + self.node2.pos) * 0.5

    @property
    def vertical(self) -> bool:
        """
        Returns true if component essentially vertical.
        """

        x1, y1 = self.node1.x, self.node1.y
        x2, y2 = self.node2.x, self.node2.y
        return abs(y2 - y1) > abs(x2 - x1)

    @property
    def label_position(self):
        """
        Returns position where to place label.
        """

        pos = self.midpoint
        w = self.label_offset
        if self.vertical:
            pos.x += w
        else:
            pos.y += w

        return pos

    def assign_positions(self, x1, y1, x2, y2) -> array:
        """Assign node positions based on cursor positions."""

        return array(((x1, y1), (x2, y2)))

    @property
    def node1(self):

        return self.nodes[0]

    @property
    def node2(self):

        return self.nodes[1]

    @property
    def angle(self):

        x1, y1 = self.node1.x, self.node1.y
        x2, y2 = self.node2.x, self.node2.y
        angle = degrees(atan2(y2 - y1, x2 - x1))
        return angle

    def attr_dir_string(self, x1, y1, x2, y2, step=1):

        r = sqrt((x1 - x2)**2 + (y1 - y2)**2) / step

        if self.type == 'X' and r >= 0.5:
            r -= 0.49

        if r == 1:
            size = ''
        else:
            size = '=' + str(round(r, 2)).rstrip('0').rstrip('.')

        angle = degrees(atan2(y2 - y1, x2 - x1)) + self.angle_offset

        if r == 0:
            attr = 'down=0'
            print('Zero length component; this will be drawn down')
        elif angle == 0:
            attr = 'right' + size
        elif angle in (90, -270):
            attr = 'up' + size
        elif angle in (180, -180):
            attr = 'left' + size
        elif angle in (270, -90):
            attr = 'down' + size
        else:
            attr = 'rotate=' + str(round(angle, 2)).rstrip('0').rstrip('.')

        return attr

    def attr_string(self, x1, y1, x2, y2, step=1):

        attr = self.attr_dir_string(x1, y1, x2, y2, step)

        if self.color != '':
            attr += ', color=' + self.color
        # TODO, add cunning way of specifing modifiers, e.g., v^, i<
        if self.voltage_label != '':
            attr += ', v=' + self.voltage_label
        if self.current_label != '':
            attr += ', i=' + self.current_label
        if self.flow_label != '':
            attr += ', f=' + self.flow_label
        if self.mirror:
            attr += ', mirror'
        if self.invert:
            attr += ', invert'

        # Add user defined attributes such as thick, dashed, etc.
        if self.attrs != '':
            attr += ', ' + self.attrs

        kind = self.symbol_kind
        if kind not in (None, ''):
            if self.type == 'X':
                attr += ', ' + kind
            else:
                attr += ', kind=' + kind

        if self.style not in (None, ''):
            attr += ', style=' + self.style

        return attr

    def attr_string_update(self, step=1):

        return self.attr_string(self.node1.x, self.node1.y,
                                self.node2.x, self.node2.y, step=step)

    def is_within_bbox(self, x, y):

        m = array((self.midpoint.x, self.midpoint.y))

        dx = self.node2.x - self.node1.x
        dy = self.node2.y - self.node1.y
        r = sqrt(dx**2 + dy**2)

        if r == 0:
            r = 0.3

        R = array(((dx, -dy), (dy, dx))) / r

        # Transform point into non-rotated box
        p = array((x, y))
        q = dot(R.T, (p - m))

        l = self.length - 0.3
        h = 0.3
        x, y = q

        # Determine if transformed point is in the box
        return x > -l / 2 and x < l / 2 and y > -h / 2 and y < h / 2

    def netitem_nodes(self, node_names):

        parts = []
        for node_name in node_names:
            parts.append(node_name)
        return parts

    @property
    def netitem_args(self):

        if self.cpt_kind == '':
            return ()
        return (self.cpt_kind, )

    def netitem(self, node_names, x1, y1, x2, y2, step=1):

        parts = [self.name]
        parts.extend(self.netitem_nodes(node_names))
        if self.type in ('E', 'G'):
            # Need to use known nodes to start with.
            parts.extend([node_names[0], node_names[1]])
        else:
            parts.extend(self.netitem_args)
        netitem = ' '.join(parts)
        attr_string = self.attr_string(x1, y1, x2, y2, step)
        netitem += '; ' + attr_string + '\n'
        return netitem

    def update(self, opts=None, nodes=None):

        if nodes is not None:
            self.nodes = nodes

        if opts is not None:
            self.opts = opts

    def choose_node_name(self, m, nodes):

        num = 1
        while True:
            name = str(num)
            if name not in nodes:
                return name
            num += 1



from tkinter import Tk, Button
from numpy import linspace
from .labelentries import LabelEntry, LabelEntries


class SubsDialog:

    def __init__(self, expr, ui, title='Substitution'):

        self.expr = expr
        self.ui = ui

        self.window = Tk()
        self.window.title(title)

        entries = []

        self.symbols = []
        for key in expr.symbols:
            # Ignore domain variable
            if key != expr.var.name:
                entries.append(LabelEntry(key, key, 0.0))
                self.symbols.append(key)

        self.labelentries = LabelEntries(self.window, ui, entries)

        button = Button(self.window, text="Subs",
                        command=self.on_update)
        button.grid(row=self.labelentries.row)

    def on_update(self):

        self.window.destroy()

        defs = {}
        for key in self.symbols:
            val = self.labelentries.get_text(key)
            if val == '':
                self.ui.show_error_dialog('Undefined symbol ' + key)
                return
            val = self.labelentries.get(key)
            defs[key] = val

        expr = self.expr.subs(defs)
        self.ui.show_expr_dialog(expr)



from tkinter import Tk, Button
from .labelentries import LabelEntry, LabelEntries


class TwoportSelectDialog:

    def __init__(self, ui, TP, model):

        self.ui = ui
        self.TP = TP
        self.window = Tk()
        self.window.title('Twoport')

        models = ('A', 'B', 'G', 'H', 'S', 'T', 'Y', 'Z')
        elements = ('matrix', '11', '12', '21', '22')

        entries = []
        entries.append(LabelEntry('model', 'Twoport model',
                                  model, models))
        entries.append(LabelEntry('element', 'Twoport element',
                                  elements[0], elements))

        self.labelentries = LabelEntries(self.window, ui, entries)

        button = Button(self.window, text="Show", command=self.on_show)
        button.grid(row=self.labelentries.row)

    def on_show(self):

        model = self.labelentries.get('model')
        element = self.labelentries.get('element')
        TP = self.TP

        if model == 'A':
            if element == 'matrix':
                expr = TP.Aparams
            elif element == '11':
                expr = TP.A11
            elif element == '12':
                expr = TP.A12
            elif element == '21':
                expr = TP.A21
            elif element == '22':
                expr = TP.A22
        elif model == 'B':
            if element == 'matrix':
                expr = TP.Bparams
            elif element == '11':
                expr = TP.B11
            elif element == '12':
                expr = TP.B12
            elif element == '21':
                expr = TP.B21
            elif element == '22':
                expr = TP.B22
        elif model == 'G':
            if element == 'matrix':
                expr = TP.Gparams
            elif element == '11':
                expr = TP.G11
            elif element == '12':
                expr = TP.G12
            elif element == '21':
                expr = TP.G21
            elif element == '22':
                expr = TP.G22
        elif model == 'H':
            if element == 'matrix':
                expr = TP.Hparams
            elif element == '11':
                expr = TP.H11
            elif element == '12':
                expr = TP.H12
            elif element == '21':
                expr = TP.H21
            elif element == '22':
                expr = TP.H22
        elif model == 'S':
            if element == 'matrix':
                expr = TP.Sparams
            elif element == '11':
                expr = TP.S11
            elif element == '12':
                expr = TP.S12
            elif element == '21':
                expr = TP.S21
            elif element == '22':
                expr = TP.S22
        elif model == 'T':
            if element == 'matrix':
                expr = TP.Tparams
            elif element == '11':
                expr = TP.T11
            elif element == '12':
                expr = TP.T12
            elif element == '21':
                expr = TP.T21
            elif element == '22':
                expr = TP.T22
        elif model == 'Y':
            if element == 'matrix':
                expr = TP.Yparams
            elif element == '11':
                expr = TP.Y11
            elif element == '12':
                expr = TP.Y12
            elif element == '21':
                expr = TP.Y21
            elif element == '22':
                expr = TP.Y22
        elif model == 'Z':
            if element == 'matrix':
                expr = TP.Zparams
            elif element == '11':
                expr = TP.Z11
            elif element == '12':
                expr = TP.Z12
            elif element == '21':
                expr = TP.Z21
            elif element == '22':
                expr = TP.Z22

        name = model + element

        self.ui.show_expr_dialog(expr, title=name)



from tkinter import Tk, Button
from .labelentries import LabelEntry, LabelEntries


class PreferencesDialog:

    def __init__(self, ui, update):

        self.model = ui.model
        self.update = update

        self.window = Tk()
        self.window.title('Preferences')

        entries = [LabelEntry('label_nodes', 'Node labels',
                              self.model.preferences.label_nodes,
                              ('all', 'none', 'alpha', 'pins',
                               'primary'), command=self.on_update),
                   LabelEntry('draw_nodes', 'Nodes',
                              self.model.preferences.draw_nodes,
                              ('all', 'none', 'connections', 'primary'),
                              command=self.on_update),
                   LabelEntry('label_cpts', 'Component labels',
                              self.model.preferences.label_cpts,
                              ('none', 'name',  'value', 'name+value'),
                              command=self.on_update),
                   LabelEntry('style', 'Style',
                              self.model.preferences.style,
                              ('american', 'british', 'european'),
                              command=self.on_update),
                   LabelEntry('voltage_dir', 'Voltage dir',
                              self.model.preferences.voltage_dir,
                              ('RP', 'EF'),
                              command=self.on_update),
                   LabelEntry('grid', 'Grid',
                              self.model.preferences.grid,
                              ('on', 'off'),
                              command=self.on_update),
                   LabelEntry('lw', 'Line width',
                              self.model.preferences.lw,
                              command=self.on_update),
                   LabelEntry('node_size', 'Node size',
                              self.model.preferences.node_size,
                              command=self.on_update),
                   LabelEntry('show_units', 'Show units',
                              self.model.preferences.show_units,
                              ('true', 'false'),
                              command=self.on_update),
                   LabelEntry('xsize', 'Width',
                              self.model.preferences.xsize,
                              command=self.on_update),
                   LabelEntry('ysize', 'Height',
                              self.model.preferences.ysize,
                              command=self.on_update),
                   LabelEntry('snap_grid', 'Snap to grid',
                              self.model.preferences.snap_grid,
                              ('true', 'false'),
                              command=self.on_update),
                   ]

        self.labelentries = LabelEntries(self.window, ui, entries)

        button = Button(self.window, text="OK", command=self.on_ok)
        button.grid(row=self.labelentries.row)

    def on_update(self, arg=None):

        self.model.preferences.label_nodes = self.labelentries.get(
            'label_nodes')
        self.model.preferences.draw_nodes = self.labelentries.get('draw_nodes')
        self.model.preferences.label_cpts = self.labelentries.get('label_cpts')
        self.model.preferences.style = self.labelentries.get('style')
        self.model.preferences.voltage_dir = self.labelentries.get(
            'voltage_dir')
        self.model.preferences.grid = self.labelentries.get('grid')
        self.model.preferences.lw = self.labelentries.get('lw')
        self.model.preferences.node_size = self.labelentries.get('node_size')
        self.model.preferences.show_units = self.labelentries.get('show_units')
        self.model.preferences.xsize = self.labelentries.get('xsize')
        self.model.preferences.ysize = self.labelentries.get('ysize')
        self.model.preferences.snap_grid = self.labelentries.get('snap_grid')

        # Do not set show_units; this needs fixing in Lcapy since
        # str(expr) includes the units and this causes problems...

        if self.update:
            # Could check for changes
            self.update()

    def on_ok(self):

        self.on_update()

        self.window.destroy()

        self.model.preferences.save()



from tkinter import Tk, Button
from .labelentries import LabelEntry, LabelEntries


class TransferFunctionDialog:

    def __init__(self, ui):

        self.ui = ui
        self.window = Tk()
        self.window.title('Transfer function')

        entries = []

        names = ui.model.circuit.cpts

        entries.append(LabelEntry('input', 'Input',
                                  names[0], names))

        entries.append(LabelEntry('output', 'Output',
                                  names[0], names))

        self.labelentries = LabelEntries(self.window, ui, entries)

        button = Button(self.window, text="OK", command=self.on_ok)
        button.grid(row=self.labelentries.row)

    def on_ok(self):

        input_cpt = self.labelentries.get('input')
        output_cpt = self.labelentries.get('output')

        print(input_cpt, output_cpt)

        self.window.destroy()



global_dict = {}
exec('from lcapy import *', global_dict)


class ExprCalc:

    def __init__(self, expr):

        self.expr = expr

    def eval(self, command):

        global_dict['result'] = self.expr

        expr = eval('(result)' + command, global_dict)
        return expr

    def attribute(self, name):

        return self.eval('.%s' % name)

    def method(self, name):

        return self.eval('.%s()' % name)



from .bipole import BipoleComponent
from math import cos, sin, radians, sqrt
from numpy import array


class Connection(BipoleComponent):

    type = 'X'
    args = ()
    has_value = False

    kinds = {'-0V': '0V',
             '-ground': 'Ground', '-sground': 'Signal ground',
             '-rground': 'Rail ground', '-cground': 'Chassis ground',
             '-vcc': 'VCC', '-vdd': 'VDD', '-vee': 'VEE', '-vss': 'VSS',
             '-input': 'Input', '-output': 'Output', '-bidir': 'Bidirectional'}

    def draw(self, editor, sketcher, **kwargs):

        x1, y1 = self.node1.x, self.node1.y
        x2, y2 = self.node2.x, self.node2.y

        kwargs = self.make_kwargs(editor, **kwargs)

        if self.symbol_kind in ('vcc', 'vdd'):
            x1, y1, angle = self.split_node_pos(x2, y2, editor.STEP, True)
            offset = x1, y1
            angle = angle + 90
        elif self.symbol_kind in ('vee', 'vss'):
            x2, y2, angle = self.split_node_pos(x1, y1, editor.STEP)
            offset = x2, y2
            angle = angle + 90
        elif self.symbol_kind in ('input', ):
            x2, y2, angle = self.split_node_pos(x1, y1, editor.STEP)
            offset = x2, y2
        else:
            x2, y2, angle = self.split_node_pos(x1, y1, editor.STEP)
            offset = x2, y2

        sketcher.sketch(self.sketch, offset=offset,
                        angle=angle, snap=False, **kwargs)

        sketcher.stroke_line(x1, y1, x2, y2, **kwargs)

    def split_node_pos(self, x, y, step=1, flip=False):

        def get_value(direction):

            val = self.opts[direction]
            if val == '':
                value = 1
            else:
                value = float(val)

            return value * step

        scale = -1 if flip else 1

        # TODO fix if positive node unknown, say for vdd, vcc.

        if 'down' in self.opts:
            angle = -90
            y -= get_value('down') * scale
        elif 'up' in self.opts:
            angle = 90
            y += get_value('up') * scale
        elif 'right' in self.opts:
            angle = 0
            x += get_value('right') * scale
        elif 'left' in self.opts:
            angle = 180
            x -= get_value('left') * scale
        elif 'angle' in self.opts:
            angle = get_value('angle')
            if 'size' in self.opts:
                size = get_value('size')
            else:
                size = 1
            x += size * cos(radians(angle))
            y += size * sin(radians(angle))
        else:
            # Assume right
            x += 1
            angle = 0

        return x, y, angle

    @property
    def sketch_net(self):

        return 'W 1 0; right=0, ' + self.symbol_kind + ', l='

    @property
    def labelled_nodes(self):

        if self.symbol_kind == '':
            return self.nodes
        elif self.symbol_kind in ('vcc', 'vdd'):
            return self.nodes[1:]
        else:
            return self.nodes[:1]

    @property
    def drawn_nodes(self):

        if self.symbol_kind == '':
            return self.nodes
        elif self.symbol_kind in ('vcc', 'vdd'):
            return self.nodes[1:]
        else:
            return self.nodes[:1]

    def is_within_bbox(self, x, y):

        # FIXME
        xm = self.midpoint.x
        ym = self.midpoint.y

        r = sqrt((x - xm)**2 + (y - ym)**2)
        return r < 0.5

    def choose_node_name(self, m, nodes):

        if m == 0 and self.symbol_kind in ('vcc', 'vdd'):
            return self.symbol_kind

        if m == 1 and self.symbol_kind in ('vee', 'vss', '0V', 'ground',
                                           'rground', 'sground', 'cground'):
            return '0'

        return super().choose_node_name(m, nodes)



from matplotlib.patches import PathPatch, Arc, Circle, Polygon
from matplotlib.transforms import Affine2D
from matplotlib.path import Path
from math import degrees
from numpy import array


class Sketcher:

    def __init__(self, ax, debug=0):

        self.ax = ax
        self.debug = debug

    def stroke_line(self, xstart, ystart, xend, yend, color='black', **kwargs):

        return self.ax.plot((xstart, xend), (ystart, yend),
                            color=color, **kwargs)

    def stroke_arc(self, x, y, r, theta1, theta2, **kwargs):

        r *= 2
        patch = Arc((x, y), r, r, 0, degrees(theta1),
                    degrees(theta2), **kwargs)
        self.ax.add_patch(patch)
        return patch

    def clear(self):

        self.ax.clear()

    def stroke_rect(self, xstart, ystart, width, height, **kwargs):
        # xstart, ystart top left corner

        xend = xstart + width
        yend = ystart + height

        self.stroke_line(xstart, ystart, xstart, yend, **kwargs)
        self.stroke_line(xstart, yend, xend, yend, **kwargs)
        self.stroke_line(xend, yend, xend, ystart, **kwargs)
        self.stroke_line(xend, ystart, xstart, ystart, **kwargs)

    def stroke_filled_circle(self, x, y, radius=0.5, color='black',
                             alpha=0.5, **kwargs):

        patch = Circle((x, y), radius, fc=color, alpha=alpha, **kwargs)
        self.ax.add_patch(patch)
        return patch

    def stroke_circle(self, x, y, radius=0.5, color='black',
                      alpha=0.5, **kwargs):

        patch = Circle((x, y), radius, fc='white',
                       color=color, alpha=alpha, **kwargs)
        self.ax.add_patch(patch)
        return patch

    def stroke_polygon(self, path, color='black', alpha=0.5,
                       fill=False, **kwargs):

        patch = Polygon(path, fc=color, alpha=alpha,
                        fill=fill, **kwargs)
        self.ax.add_patch(patch)
        return patch

    def text(self, x, y, text, **kwargs):

        from lcapy.latex import latex_format_label

        # The matplotlib mathtext parser does not like
        # dollar signs inside mathrm, e.g., \mathrm{$A_2$}
        # text = r'$\mathrm{' + latex_format_label(text) + '}$'

        return self.ax.annotate(text, (x, y), **kwargs)

    def stroke_path(self, path, color='black', **kwargs):

        for m in range(len(path) - 1):
            xstart, ystart = path[m]
            xend, yend = path[m + 1]

            self.stroke_line(xstart, ystart, xend, yend, color=color, **kwargs)

    def remove(self, patch):

        self.ax.remove(patch)

    def sketch(self, sketch, offset=(0, 0), scale=1, angle=0, **kwargs):

        kwargs = {**sketch.kwargs, **kwargs}

        gtransform = Affine2D().rotate_deg(angle).scale(scale * sketch.SCALE)
        gtransform = gtransform.translate(*offset)

        color = kwargs.pop('color', sketch.color)
        mirror = kwargs.pop('mirror', False)
        invert = kwargs.pop('invert', False)

        patches = []

        for m, spath in enumerate(sketch.paths):
            path = spath.path
            fill = spath.fill

            # Note, the SVG coordinate system has y going down the screen
            # but Matplotlib's coordinate system has y going up the screen.
            # Thus we need to invert the sense of mirror.

            if not mirror:
                vertices = path.vertices * (1, -1)
                path = Path(vertices, path.codes)
            if invert:
                vertices = path.vertices * (-1, 1)
                path = Path(vertices, path.codes)

            path = path.transformed(gtransform)

            if False and patches == []:
                print(path.vertices)

            fill = kwargs.pop('fill', fill)

            if self.debug:
                color = ['red', 'yellow', 'orange',
                         'green', 'blue', 'violet'][m % 6]

            patch = PathPatch(path, fill=fill, color=color, **kwargs)
            patches.append(patch)
            self.ax.add_patch(patch)

        return patches



from tkinter import Tk, Button
from numpy import linspace
from .labelentries import LabelEntry, LabelEntries


class LimitDialog:

    def __init__(self, expr, ui, title='Limit'):

        self.expr = expr
        self.ui = ui

        self.window = Tk()
        self.window.title(title)

        symbols = list(expr.symbols)
        if len(symbols) == 0:
            raise ValueError('Constant expression')

        entries = []
        entries.append(LabelEntry('symbol', 'symbol',
                                  symbols[0], symbols))

        entries.append(LabelEntry('limit', 'Limit', 0.0))

        entries.append(LabelEntry('dir', 'Direction',
                                  '+', ('+', '-')))

        self.labelentries = LabelEntries(self.window, ui, entries)

        button = Button(self.window, text="OK", command=self.on_ok)
        button.grid(row=self.labelentries.row)

    def on_ok(self):

        self.window.destroy()

        symbol = self.labelentries.get('symbol')
        limit = self.labelentries.get('limit')
        dir = self.labelentries.get('dir')

        expr = self.expr.limit(symbol, limit, dir)
        self.ui.show_expr_dialog(expr)



from .bipole import BipoleComponent


class Capacitor(BipoleComponent):

    type = 'C'
    default_kind = '-'
    kinds = {'-': '', '-electrolytic': 'Electrolytic',
             '-polar': 'Polar', '-variable': 'Variable',
             '-curved': 'Curved', '-sensor': ' Sensor',
             '-tunable': 'Tunable'}



from tkinter import StringVar, Label, Entry, OptionMenu
from tkinter.ttk import Checkbutton


class LabelEntry:

    def __init__(self, name, text, default, options=None,
                 command=None, **kwargs):

        self.name = name
        self.text = text
        if default is None:
            default = 'None'
        self.default = default
        self.cls = default.__class__
        self.options = options
        self.command = command
        # width, etc.
        self.kwargs = kwargs


class LabelEntries(dict):

    def __init__(self, window, ui, entries):

        self.row = 0
        self.ui = ui

        for labelentry in entries:

            var = StringVar(window)
            var.set(labelentry.default)
            self[labelentry.name] = (var, labelentry.cls)

            label = Label(window, text=labelentry.text + ': ', anchor='w')
            if isinstance(labelentry.options, (tuple, list)):
                entry = OptionMenu(
                    window, var, *labelentry.options,
                    command=labelentry.command)
            else:
                if isinstance(labelentry.default, bool):
                    entry = Checkbutton(
                        window, variable=var, command=labelentry.command)
                else:
                    entry = Entry(window, textvariable=var,
                                  **labelentry.kwargs)

                    # if labelentry.command:
                    #    var.trace_add('write', labelentry.command)
                    if labelentry.command:
                        entry.bind('<Return>', labelentry.command)

            label.grid(row=self.row, sticky='w')
            entry.grid(row=self.row, column=1, sticky='w')

            self.row += 1

    def get_var(self, name):

        return self[name][0]

    def get_cls(self, name):

        return self[name][1]

    def get_text(self, name):

        val = self.get_var(name).get()
        if val == 'None':
            val = None
        return val

    def get(self, name):

        val = self.get_text(name)
        if val is None:
            return val

        cls = self.get_cls(name)

        if cls is bool:
            val = {'1': True, '0': False}[val]

        try:
            return cls(val)
        except Exception:
            self.ui.show_error_dialog('Invalid value %s for %s' % (val, name))



from tkinter import Menu


class MenuItem:

    def __init__(self, label, command=None, underline=0, accelerator=None):

        self.label = label
        self.command = command
        self.underline = underline
        self.accelerator = accelerator


class MenuDropdown:

    def __init__(self, label, underline=0, menuitems=None):

        self.label = label
        self.underline = underline
        self.menuitems = menuitems


class MenuBar:

    def __init__(self, menudropdowns):

        self.menudropdowns = menudropdowns

    def make(self, window):

        def doit(menuitem):

            menuitem.command(menuitem.label)

        # Create the drop down menus
        self.menubar = Menu(window, bg='lightgrey', fg='black')

        self.menus = []

        for menudropdown in self.menudropdowns:
            menu = Menu(self.menubar, tearoff=0, bg='lightgrey', fg='black')

            self.menubar.add_cascade(label=menudropdown.label,
                                     underline=menudropdown.underline,
                                     menu=menu)

            for menuitem in menudropdown.menuitems:

                if menuitem is None:
                    continue

                if isinstance(menuitem, MenuDropdown):

                    submenu = Menu(self.menubar, tearoff=0,
                                   bg='lightgrey', fg='black')
                    menu.add_cascade(label=menuitem.label,
                                     underline=menuitem.underline,
                                     menu=submenu)
                    for submenuitem in menuitem.menuitems:
                        submenu.add_command(label=submenuitem.label,
                                            command=lambda a=submenuitem: doit(
                                                a),
                                            underline=submenuitem.underline,
                                            accelerator=submenuitem.accelerator)

                else:
                    menu.add_command(label=menuitem.label,
                                     command=lambda a=menuitem: doit(a),
                                     underline=menuitem.underline,
                                     accelerator=menuitem.accelerator)

            self.menus.append(menu)

        window.config(menu=self.menubar)



from .components.port import Port
from numpy import array


class Node:

    def __init__(self, x, y, name):

        self.x = float(x)
        self.y = float(y)
        self.name = name
        self.count = 0
        self.cpts = []

    @property
    def position(self):

        return array((self.x, self.y))

    @position.setter
    def position(self, pos):

        self.x = pos[0]
        self.y = pos[1]

    def __str__(self):

        # Note, need to have float type otherwise 0 becomes empty string.
        x = str(round(self.x, 2)).rstrip('0').rstrip('.')
        y = str(round(self.y, 2)).rstrip('0').rstrip('.')

        return '%s@(%s, %s)' % (self.name, x, y)

    @property
    def is_primary(self):

        name = self.name
        parts = name.split('_')
        return (name[0] != '_' and len(parts) <= 2) \
            and not (name[0].isdigit() and len(parts) != 1)

    @property
    def port(self):

        for cpt in self.cpts:
            if isinstance(cpt, Port):
                return True
        return False

    def debug(self):

        s = str(self) + ', count=%s' % self.count

        names = [cpt.name for cpt in self.cpts]

        s += ', cpts=[%s]' % ', '.join(names) + '\n'

        return s



from tkinter import Tk, StringVar, Label, Entry, Button


class NodePropertiesDialog:

    def __init__(self, node, update=None, title=''):

        self.node = node
        self.update = update

        self.window = Tk()
        self.window.title(title)

        row = 0

        self.name_var = StringVar(self.window)
        self.name_var.set(node.name)

        name_label = Label(self.window, text='Name: ')
        name_entry = Entry(self.window, textvariable=self.name_var)

        name_label.grid(row=row)
        name_entry.grid(row=row, column=1)
        row += 1

        button = Button(self.window, text="OK", command=self.on_update)
        button.grid(row=row)

    def on_update(self):

        node_name = self.name_var.get()

        self.window.destroy()

        if self.node.name != node_name and self.update:
            self.node.name = node_name
            self.update(self.node)



from .node import Node


class Nodes(list):

    def __init__(self):

        super(Nodes, self).__init__(self)

    def add(self, node):

        if not isinstance(node, Node):
            raise TypeError('Not a Node object')

        if node not in self:
            self.append(node)

    def by_name(self, name):

        for node in self:
            if node.name == name:
                return node
        return None

    def by_position(self, position):

        x, y = position

        for node in self:
            if node.x == x and node.y == y:
                return node
        return None

    def clear(self):

        while self != []:
            self.pop()

    def closest(self, x, y):

        for node in self:
            x1, y1 = node.position
            rsq = (x1 - x)**2 + (y1 - y)**2
            if rsq < 0.1:
                return node
        return None

    def debug(self):

        s = ''
        for node in self:
            s += node.debug()
        return s

    def make(self, x, y, name, cpt):

        node = self.by_position((x, y))
        if node is not None:
            if name is not None and node.name != name:
                raise ValueError('Node name conflict %s and %s' %
                                 (node.name, name))
            node.count += 1
            node.cpts.append(cpt)
            return node

        if name is None:
            num = 1
            while True:
                name = str(num)
                if not self.by_name(name):
                    break
                num += 1

        node = Node(x, y, name)
        node.count += 1
        node.cpts.append(cpt)
        return node

    def remove(self, node, cpt):

        node.count -= 1
        if node.count == 0:
            self.pop(self.index(node))

        node.cpts.pop(node.cpts.index(cpt))



class Annotation:

    def __init__(self, ui, x, y, text):

        self.sketcher = ui.sketcher
        self.x = x
        self.y = y
        self.text = text
        self.patch = None

    @property
    def position(self):

        return self.x, self.y

    def draw(self, **kwargs):

        self.patch = self.sketcher.text(self.x, self.y, self.text, **kwargs)

    def remove(self):

        if self.patch:
            self.patch.remove()



from lcapy import Circuit
from matplotlib.transforms import Affine2D
from .svgparse import SVGParse
from os.path import join


class SketchPath:

    def __init__(self, path, style, symbol):

        self.path = path
        self.style = style
        self.symbol = symbol

    @property
    def fill(self):

        return self.symbol or ('fill' in self.style and self.style['fill'] != 'none')

    def transform(self, transform):

        path = self.path.transformed(transform)

        return self.__class__(path, self.style, self.symbol)


class Sketch:

    # Convert points to cm.
    SCALE = 2.54 / 72

    def __init__(self, paths, width, height, **kwargs):

        self.paths = paths
        self.width = width
        self.height = height
        self.kwargs = kwargs

    @property
    def color(self):

        return self.kwargs.get('color', 'black')

    @classmethod
    def load(cls, sketch_key):

        from lcapygui import __datadir__

        dirname = __datadir__ / 'svg'
        svg_filename = dirname / (sketch_key + '.svg')

        if not svg_filename.exists():
            return None

        svg = SVGParse(str(svg_filename))

        sketch_paths = []
        for svga_path in svg.paths:
            sketch_path = SketchPath(
                svga_path.path, svga_path.style, svga_path.symbol)
            sketch_path = sketch_path.transform(Affine2D(svga_path.transform))
            sketch_paths.append(sketch_path)

        sketch = cls(sketch_paths, svg.width, svg.height).align(sketch_key)
        return sketch

    @classmethod
    def create(cls, sketch_key, sketch_net):

        dirname = join('lcapygui', 'data', 'svg')
        svg_filename = join(dirname, sketch_key + '.svg')

        a = Circuit()

        net = sketch_net
        if net is None:
            return None
        if ';' not in net:
            net += '; right'

        a.add(net)

        a.draw(str(svg_filename), label_values=False, label_ids=False,
               label_nodes=False, draw_nodes=False)

    def offsets1(self, sketch_key):

        if sketch_key == 'opamp':
            return 0, self.height / 2

        # TODO, use sketch_key to help find offset.

        xoffset = None
        yoffset = None

        # Look for pair of horizontal wires
        candidates = []
        for path in self.paths:
            if len(path.path) >= 4 and all(path.path.codes[0:4] == (1, 2, 1, 2)):
                vertices = path.path.vertices
                if vertices[0][1] == vertices[1][1]:
                    xoffset = vertices[0][0]
                    yoffset = vertices[0][1]
                    candidates.append((xoffset, yoffset))

        if candidates != []:
            # Search for horizontal line with longest extent.
            # This is needed for fet/bjt.
            xmin = 1000
            yoffset = 0
            for candidate in candidates:
                if candidate[0] < xmin:
                    xmin = candidate[0]
                    yoffset = candidate[1]

            return 0, yoffset

        # Look for vertical wire (for ground, sground, cground, rground)
        # Note, if look for horizontal wire first, get incorrect offset for rground
        for path in self.paths:
            if len(path.path) == 2 and all(path.path.codes == (1, 2)):
                vertices = path.path.vertices
                if vertices[0][0] == vertices[1][0]:
                    xoffset = vertices[0][0]
                    yoffset = vertices[0][1]
                    return xoffset, yoffset

        # Look for single horizontal wire (this is triggered by W components)
        for path in self.paths:
            if len(path.path) == 2 and all(path.path.codes == (1, 2)):
                vertices = path.path.vertices
                if vertices[0][1] == vertices[1][1]:
                    xoffset = vertices[0][0]
                    yoffset = vertices[0][1]
                    return 0, yoffset

        return self.width / 2, self.height / 2

    def offsets(self, sketch_key):

        xoffset, yoffset = self.offsets1(sketch_key)
        return xoffset, yoffset

    def align(self, sketch_key):
        """Remove yoffset from component."""

        xoffset, yoffset = self.offsets(sketch_key)

        if xoffset is None:
            return self

        paths = []
        for path in self.paths:
            paths.append(path.transform(
                Affine2D().translate(-xoffset, -yoffset)))

        return self.__class__(paths, self.width, self.height, **self.kwargs)



from numpy import arange


class Drawing():

    def __init__(self, ui, fig, model, debug=0):

        self.ui = ui
        self.fig = fig
        self.debug = debug
        self.xsize = model.preferences.xsize
        self.ysize = model.preferences.ysize

        self.ax = self.fig.add_subplot(111)

        self.draw_grid('on')
        self.set_default_view()

    def draw_grid(self, grid):

        if self.debug:
            print('draw grid')

        # Enlarge grid by factor of 2 in each direction.
        # Only xsize by ysize is visible.
        xticks = arange(self.xsize * 2)
        yticks = arange(self.ysize * 2)

        self.ax.axis('equal')
        self.ax.set_xticks(xticks)
        self.ax.set_yticks(yticks)
        self.ax.set_xticklabels([])
        self.ax.set_yticklabels([])
        if grid == 'on':
            self.ax.grid(color='lightblue')

        self.ax.tick_params(which='both', left=False, bottom=False,
                            top=False, labelbottom=False)

        self.ax.set_axisbelow(True)

    def savefig(self, filename):

        self.fig.savefig(filename, bbox_inches='tight', pad_inches=0)

    def set_view(self, xmin, ymin, xmax, ymax):

        if self.debug:
            print('view', xmin, ymin, xmax, ymax)

        self.ax.set_xlim(xmin, xmax)
        self.ax.set_ylim(ymin, ymax)

    def set_default_view(self):

        self.set_view(0, 0, self.xsize, self.ysize)

    def clear(self, grid='on'):

        if self.debug:
            print('clear')

        xmin, xmax = self.ax.get_xlim()
        ymin, ymax = self.ax.get_ylim()

        self.ax.clear()

        self.draw_grid(grid)

        self.set_view(xmin, ymin, xmax, ymax)

    def refresh(self):

        if self.debug:
            print('refresh')
        self.fig.canvas.draw()



from tkinter import Tk, Entry, Button, Text, BOTH, END
from lcapy import expr
from re import sub

global_dict = {}
exec('from lcapy import *', global_dict)


def exec_function(expr):

    lines = expr.split('\n')
    lines[-1] = 'return ' + lines[-1]
    s = '\n'.join(lines)

    context = {}
    local_vars = {}

    s = sub(r"(?m)^", "    ", s)
    s = "def _anon():\n" + s

    exec(s, global_dict, context)
    result = context['_anon']()

    return result


class PythonDialog:

    def __init__(self, expr, ui):

        self.expr = expr
        self.ui = ui

        self.window = Tk()
        self.window.title('Expression editor')

        symbols = self.expr.symbols

        s = ''
        for sym in symbols:
            # Skip domain variables
            if sym in ('f', 's', 't', 'w', 'omega',
                       'jf', 'jw', 'jomega', 'n', 'k', 'z'):
                continue

            # TODO, add other assumptions
            if symbols[sym].is_positive:
                s += "%s = symbol('%s', positive=True)\n" % (sym, sym)
            else:
                s += "%s = symbol('%s')\n" % (sym, sym)

        s += repr(self.expr)

        self.text = Text(self.window)
        self.text.pack(fill=BOTH, expand=1)
        self.text.insert(END, s)

        button = Button(self.window, text='Show', command=self.on_show)
        button.pack()

    def on_show(self):

        expr_str = self.text.get('1.0', END).strip()

        try:
            expr = exec_function(expr_str)
            self.ui.show_expr_dialog(expr)
            self.window.destroy()

        except Exception as e:
            self.ui.show_error_dialog('Cannot evaluate expression')



from .bipole import BipoleComponent


class CurrentSource(BipoleComponent):

    type = 'I'
    kinds = {'dc': 'DC', 'ac': 'AC', 'step': 'Step',
             '': 'Arbitrary', 'noise': 'Noise', 's': ''}
    default_kind = ''

    @property
    def sketch_net(self):

        return self.type + ' 1 2 ' + self.kind + ' ' + '; right'



from .uimodelbase import UIModelBase
from lcapy.mnacpts import Cpt
from lcapy import Circuit
from os.path import basename
from warnings import warn


class Cursor:

    def __init__(self, ui, x, y):

        self.sketcher = ui.sketcher
        self.patch = None
        self.x = x
        self.y = y

    @property
    def position(self):

        return self.x, self.y

    def draw(self, color='red', radius=0.3):

        self.patch = self.sketcher.stroke_filled_circle(self.x, self.y,
                                                        radius,
                                                        color=color,
                                                        alpha=0.5)

    def remove(self):

        self.patch.remove()


class Cursors(list):

    def debug(self):

        s = ''
        for cursor in self:
            s += '%s, %s' % (cursor.x, cursor.y) + '\n'
        return s

    def remove(self):

        while self != []:
            self.pop().remove()

    def draw(self):

        if len(self) > 0:
            self[0].draw('red')
        if len(self) > 1:
            self[1].draw('blue')


class UIModelMPH(UIModelBase):

    def __init__(self, ui):

        super(UIModelMPH, self).__init__(ui)

        self.cursors = Cursors()
        self.node_cursor = None

        self.key_bindings = {
            'ctrl+c': self.on_copy,
            'ctrl+d': self.on_debug,
            'ctrl+e': self.on_export,
            'ctrl+h': self.on_help,
            'ctrl+i': self.on_inspect,
            'ctrl+n': self.on_new,
            'ctrl+o': self.on_load,
            'ctrl+s': self.on_save,
            'alt+s': self.on_save_as,
            'ctrl+u': self.on_view,
            'ctrl+v': self.on_paste,
            'ctrl+w': self.on_quit,
            'ctrl+x': self.on_cut,
            'ctrl+y': self.on_redo,
            'ctrl+z': self.on_undo,
            'ctrl+9': self.on_pdb,
            'escape': self.on_unselect,
            'delete': self.on_delete,
            'backspace': self.on_delete}

        self.key_bindings_with_key = {}
        for k, v in self.component_map.items():
            self.key_bindings_with_key[k] = self.on_add_cpt
        for k, v in self.connection_map.items():
            self.key_bindings_with_key[k] = self.on_add_con

    def add_cursor(self, x, y):

        cursor = Cursor(self.ui, x, y)

        if len(self.cursors) == 0:
            cursor.draw('red')
            self.cursors.append(cursor)

        elif len(self.cursors) == 1:
            cursor.draw('blue')
            self.cursors.append(cursor)

        elif len(self.cursors) == 2:

            rp = (x - self.cursors[0].x)**2 + (y - self.cursors[0].y)**2
            rm = (x - self.cursors[1].x)**2 + (y - self.cursors[1].y)**2

            if rm > rp:
                # Close to plus cursor so add new minus cursor
                self.cursors[1].remove()
                self.cursors[1] = cursor
                self.cursors[1].draw('blue')
            else:
                # Close to minus cursor so change minus cursor to plus cursor
                # and add new minus cursor
                self.cursors[0].remove()
                self.cursors[1].remove()
                self.cursors[0] = self.cursors[1]
                self.cursors[0].draw('red')
                self.cursors[1] = cursor
                self.cursors[1].draw('blue')

        self.ui.refresh()

    def clear(self):

        self.ui.clear(self.preferences.grid)

    def closest_cpt(self, x, y):

        for cpt in self.circuit.elements.values():

            gcpt = cpt.gcpt
            if gcpt is None:
                continue

            if gcpt.is_within_bbox(x, y):
                return cpt

        return None

    def closest_node(self, x, y):

        for node in self.circuit.nodes.values():
            if node.pos is None:
                # This happens with opamps.  Node 0 is the default
                # reference pin.
                warn('Ignoring node %s with no position' % node.name)
                continue
            x1, y1 = node.pos.x, node.pos.y
            rsq = (x1 - x)**2 + (y1 - y)**2
            if rsq < 0.1:
                return node
        return None

    def create_state_space(self, cpt):

        ss = self.circuit.ss
        self.ui.show_state_space_dialog(ss)

    def create_transfer_function(self, cpt):

        self.ui.show_transfer_function_dialog(cpt)

    def create_twoport(self, cpt, kind):

        self.ui.show_twoport_dialog(cpt, kind)

    def exception(self, e):
        message = str(e)
        if self.pathname != '':
            message += ' in ' + self.pathname
        self.ui.show_error_dialog(message)
        if self.ui.debug:
            import pdb
            pdb.set_trace()

    def new_name(self, pathname):

        from os.path import splitext

        base, ext = splitext(pathname)
        parts = base.split('_')
        if len(parts) == 0:
            suffix = '1'
        else:
            try:
                suffix = str(int(parts[-1]) + 1)
                base = '_'.join(parts[0:-1])
            except ValueError:
                suffix = '1'
        return base + '_' + suffix + ext

    def on_add_node(self, x, y):

        # Snap to known node then snap to grid.
        node = self.closest_node(x, y)
        if node is None:
            if self.preferences.snap_grid == 'true':
                x, y = self.snap(x, y)
            # Could be smarter and try to align with nearby node.
        else:
            x, y = node.x, node.y

        self.add_cursor(x, y)

    def on_add_cpt(self, cpt_key):

        if self.ui.debug:
            print(cpt_key)

        if len(self.cursors) == 0:
            self.ui.show_info_dialog(
                'To add component, first create nodes by clicking on grid')
            return
        elif len(self.cursors) == 1:
            self.ui.show_info_dialog(
                'To add component, add negative node by clicking on grid')
            return

        x1 = self.cursors[0].x
        y1 = self.cursors[0].y
        x2 = self.cursors[1].x
        y2 = self.cursors[1].y

        self.cpt_create(cpt_key, x1, y1, x2, y2)
        self.ui.refresh()

    def on_add_con(self, con_key):

        if self.ui.debug:
            print(con_key)

        if len(self.cursors) == 0:
            self.ui.show_info_dialog(
                'To add component, first create nodes by clicking on grid')
            return
        elif len(self.cursors) == 1:
            self.ui.show_info_dialog(
                'To add component, add negative node by clicking on grid')
            return

        # TODO: if have a single cursor choose down direction.

        x1 = self.cursors[0].x
        y1 = self.cursors[0].y
        x2 = self.cursors[1].x
        y2 = self.cursors[1].y

        self.con_create(con_key, x1, y1, x2, y2)
        self.ui.refresh()

    def on_best_fit(self):

        bbox = self.bounding_box()
        if bbox is None:
            return
        xmin, ymin, xmax, ymax = bbox

        self.ui.set_view(xmin - 2, ymin - 2, xmax + 2, ymax + 2)
        self.ui.refresh()

    def on_clone(self):

        pathname = self.new_name(self.pathname)
        self.save(pathname)

        model = self.ui.new()
        model.load(pathname)
        filename = basename(pathname)
        self.ui.set_filename(filename)
        self.ui.refresh()

    def on_close(self):

        self.ui.quit()

    def on_copy(self):

        if self.selected is None:
            return
        if not self.cpt_selected:
            return

        self.copy(self.selected)

    def on_cpt_changed(self, cpt):

        self.invalidate()
        # Component name may have changed
        self.clear()

        if isinstance(cpt, Cpt):

            # If kind has changed need to remake the sketch
            # and remake the cpt.
            # If name changed need to remake the cpt.
            self.cpt_remake(cpt)
        else:
            # Node name may have changed...
            pass

        self.redraw()
        self.cursors.draw()
        self.ui.refresh()

    def on_create_state_space(self):

        self.create_state_space(self.selected)

    def on_create_transfer_function(self):

        self.create_transfer_function(self.selected)

    def on_create_twoport(self, kind):

        self.create_twoport(self.selected, kind)

    def on_cut(self):

        if self.selected is None:
            return
        if not self.cpt_selected:
            return

        self.cut(self.selected)

        self.cursors.draw()

        self.ui.refresh()

    def on_debug(self):

        s = ''
        s += 'Netlist.........\n'
        s += self.schematic() + '\n'
        s += 'Nodes...........\n'
        s += self.circuit.nodes.debug() + '\n'
        s += 'Cursors.........\n'
        s += self.cursors.debug() + '\n'
        s += 'Selected.........\n'
        s += str(self.selected) + '\n'
        self.ui.show_message_dialog(s, 'Debug')

    def on_delete(self):

        if self.selected is None:
            return
        if not self.cpt_selected:
            # Handle node deletion later
            return

        self.delete(self.selected)

        self.cursors.draw()

        self.ui.refresh()

    def on_describe(self):

        self.ui.show_message_dialog(self.circuit.description(),
                                    title='Description')

    def on_expand(self):

        cct = self.circuit.expand()
        self.on_show_new_circuit(cct)

    def on_export(self):

        pathname = self.ui.export_file_dialog(self.pathname)
        if pathname == '':
            return
        self.export(pathname)

    def on_expression(self):

        from lcapy import expr

        e = self.last_expr if self.last_expr is not None else expr(0)
        self.ui.show_expr_dialog(e)

    def on_help(self):

        self.ui.show_help_dialog()

    def on_inspect(self):

        if not self.selected:
            return

        if not self.cpt_selected:
            return

        self.ui.show_inspect_dialog(self.selected,
                                    title=self.selected.name)

    def on_inspect_current(self):

        if not self.selected or not self.cpt_selected:
            return

        win = self.ui.show_working_dialog('Calculating voltage')
        self.inspect_current(self.selected)
        win.destroy()

    def on_inspect_norton_admittance(self):

        if not self.selected or not self.cpt_selected:
            return

        self.inspect_norton_admittance(self.selected)

    def on_inspect_thevenin_impedance(self):

        if not self.selected or not self.cpt_selected:
            return

        self.inspect_thevenin_impedance(self.selected)

    def on_inspect_voltage(self):

        if not self.selected or not self.cpt_selected:
            return

        win = self.ui.show_working_dialog('Calculating voltage')
        self.inspect_voltage(self.selected)
        win.destroy()

    def on_laplace_model(self):

        cct = self.circuit.s_model()
        self.on_show_new_circuit(cct)

    def on_left_click(self, x, y):

        self.on_select(x, y)

        if self.cpt_selected:
            cpt = self.selected
            if self.ui.debug:
                print('Selected ' + cpt.name)
            self.cursors.remove()
            self.add_cursor(cpt.gcpt.node1.pos.x, cpt.gcpt.node1.pos.y)
            node2 = cpt.gcpt.node2
            if node2 is not None:
                self.add_cursor(node2.pos.x, node2.pos.y)
        else:
            if self.ui.debug:
                print('Add node at (%s, %s)' % (x, y))
            self.on_add_node(x, y)

    def on_left_double_click(self, x, y):

        self.on_right_click(x, y)

    def on_load(self, initial_dir='.'):

        pathname = self.ui.open_file_dialog(initial_dir)
        if pathname == '' or pathname == ():
            return

        model = self.ui.new()
        model.load(pathname)
        self.ui.set_filename(pathname)
        self.ui.refresh()

    def on_manipulation_kill(self):

        # Could have a dialog to select what to kill

        cct = self.circuit.kill()
        self.on_show_new_circuit(cct)

    def on_manipulation_remove_sources(self):

        # Could have a dialog to select what to remove

        # Remove independent sources
        cct = self.circuit.copy()
        cct = cct.copy()
        values = list(cct.elements.values())
        for cpt in values:
            if cpt.is_independent_source:
                cct.remove(cpt.name)

        self.on_show_new_circuit(cct)

    def on_mesh_equations(self):

        if self.ground_node is None:
            self.ui.show_info_dialog('Suggest adding a ground node.')

        try:
            la = self.analysis_circuit.loop_analysis()
        except Exception as e:
            self.exception(e)
            return

        eqns = la.mesh_equations()
        self.ui.show_equations_dialog(eqns, 'Mesh equations')

    def on_move(self, xshift, yshift):

        self.move(xshift, yshift)

    def on_simple_netlist(self):

        netlist = []
        lines = self.circuit.netlist().split('\n')
        for line in lines:
            parts = line.split(';')
            netlist.append(parts[0].strip())
        s = '\n'.join(netlist)
        self.ui.show_message_dialog(s, 'Netlist')

    def on_netlist(self):

        s = self.schematic()
        self.ui.show_message_dialog(s, 'Netlist')

    def on_nodal_equations(self):

        if self.ground_node is None:
            self.ui.show_info_dialog('Suggest adding a ground node.')

        try:
            na = self.analysis_circuit.nodal_analysis()
        except Exception as e:
            self.exception(e)
            return

        eqns = na.nodal_equations()
        self.ui.show_equations_dialog(eqns, 'Nodal equations')

    def on_new(self):

        self.ui.new()

    def on_noise_model(self):

        cct = self.circuit.noise_model()
        self.on_show_new_circuit(cct)

    def on_paste(self):

        if len(self.cursors) < 2:
            # TODO, place cpt where mouse is...
            return
        x1 = self.cursors[0].x
        y1 = self.cursors[0].y
        x2 = self.cursors[1].x
        y2 = self.cursors[1].y

        self.paste(x1, y1, x2, y2)
        self.ui.refresh()

    def on_pdb(self):

        import pdb
        pdb.set_trace()

    def on_preferences(self):

        self.ui.show_preferences_dialog(self.on_redraw)

    def on_quit(self):

        if self.dirty:
            self.ui.show_info_dialog('Schematic not saved')
        else:
            self.ui.quit()

    def on_redo(self):

        self.redo()
        self.ui.refresh()

    def on_redraw(self):

        self.clear()
        self.redraw()
        self.ui.refresh()

    def on_right_click(self, x, y):

        self.on_select(x, y)
        if not self.selected:
            return

        if self.cpt_selected:
            self.ui.inspect_properties_dialog(self.selected,
                                              self.on_cpt_changed,
                                              title=self.selected.name)
        else:
            self.ui.show_node_properties_dialog(self.selected,
                                                self.on_cpt_changed,
                                                title='Node ' +
                                                self.selected.name)

    def on_right_double_click(self, x, y):
        pass

    def on_rotate(self, angle):

        self.rotate(angle)

    def on_save(self):

        pathname = self.pathname
        if pathname == '':
            return
        self.save(pathname)
        self.ui.save(pathname)

    def on_save_as(self):

        pathname = self.ui.save_file_dialog(self.pathname)
        if pathname == '' or pathname == ():
            return
        self.save(pathname)
        self.ui.save(pathname)

    def on_screenshot(self):

        pathname = self.ui.export_file_dialog(self.pathname,
                                              default_ext='.png')
        if pathname == '' or pathname == ():
            return
        self.ui.screenshot(pathname)

    def on_select(self, x, y):

        cpt = self.closest_cpt(x, y)

        if cpt is None:
            node = self.closest_node(x, y)

        if cpt:
            self.select(cpt)
        elif node:
            self.select(node)
        else:
            self.select(None)

    def on_show_new_circuit(self, cct):

        model = self.ui.new()
        model.load_from_circuit(cct)

        pathname = self.new_name(self.pathname)
        filename = basename(pathname)
        self.ui.set_filename(filename)
        self.ui.refresh()

    def on_undo(self):

        self.undo()
        self.ui.refresh()

    def on_unselect(self):

        self.unselect()

    def on_view(self):

        self.view()

    def on_view_macros(self):

        from lcapy.system import tmpfilename
        from os import remove

        schtex_filename = tmpfilename('.schtex')

        cct = Circuit(self.schematic())
        cct.draw(schtex_filename)

        with open(schtex_filename) as f:
            content = f.read()
        remove(schtex_filename)

        self.ui.show_message_dialog(content)

    def unselect(self):

        self.cursors.remove()
        self.ui.refresh()



from tkinter import Tk
from tkinter.messagebox import showinfo


class WorkingDialog:

    def __init__(self, message, title=''):

        if False:
            # Need to make this non-modal
            self.window = Tk()
            self.window.title(title)

            showinfo('', message)

        # Marvellous if could display warning messages here...

    def destroy(self):

        if False:
            self.window.destroy()



from tkinter import Tk
from .menu import MenuBar, MenuDropdown, MenuItem
from .labelentries import LabelEntry, LabelEntries


class StateSpaceDialog(Tk):

    def __init__(self, ui, ss):

        super().__init__()

        self.ui = ui
        self.title('State space')
        self.ss = ss
        self.kindmap = {'State equations': 'state_equations',
                        'Output equations': 'output_equations',
                        'State matrix, A': 'A',
                        'Input matrix, B': 'B',
                        'Output matrix, C': 'C',
                        'Feed through matrix, D': 'D',
                        'Transfer function vector, G': 'G',
                        'System transfer functions matrix, H': 'H',
                        'Eigenvalue matrix, Lambda': 'Lambda',
                        'Modal matrix, M': 'M',
                        'Characteristic polynomial': 'P',
                        'State transition matrix (Laplace), Phi': 'Phi',
                        'Input vector (Laplace), U': 'U',
                        'State vector (Laplace), X': 'X',
                        'Output vector (Laplace), Y': 'Y',
                        'Input vector, u': 'u',
                        'State vector, x': 'x',
                        'State initial value vector, x0': 'x0',
                        'Output vector, y': 'y',
                        'System impulse responses matrix, h': 'h',
                        'Controllability matrix': 'controllability_matrix',
                        'Observability matrix': 'observability_matrix'}

        items = []
        for v in self.kindmap:
            items.append(MenuItem(v, self.on_view))

        mdd = MenuDropdown('View', 0, items)

        menudropdowns = [
            mdd,
            MenuDropdown('Manipulate', 0,
                         [MenuItem('Discretize', self.on_discretize)
                          ])]

        self.menubar = MenuBar(menudropdowns)
        self.menubar.make(self)

        self.minsize(200, 20)

    def on_view(self, arg):

        attr = getattr(self.ss, self.kindmap[arg])
        try:
            expr = attr()
        except (AttributeError, TypeError):
            expr = attr

        self.ui.show_expr_dialog(expr)

    def on_discretize(self, arg):

        from lcapy.dtstatespace import DTStateSpace

        if isinstance(arg, DTStateSpace):
            return

        ssd = self.ss.discretize()
        self.ui.show_state_space_dialog(ssd)



from .component import Component
from numpy import array, sqrt, nan
from numpy.linalg import norm


def point_in_triangle(x, y, x0, y0, x1, y1, x2, y2):

    s = (x0 - x2) * (y - y2) - (y0 - y2) * (x - x2)
    t = (x1 - x0) * (y - y0) - (y1 - y0) * (x - x0)

    if ((s < 0) != (t < 0) and s != 0 and t != 0):
        return False

    d = (x2 - x1) * (y - y1) - (y2 - y1) * (x - x1)

    return d == 0 or (d < 0) == (s + t <= 0)


class Opamp(Component):

    type = "Eopamp"
    sketch_net = 'E 1 2 opamp 3 4'
    sketch_key = 'opamp'
    label_offset = 0
    args = ('Ad', 'Ac', 'Ro')

    # The Nm node is not used (ground).
    node_pinnames = ('out', '', 'in+', 'in-')

    ppins = {'out': ('rx', 1.25, 0.0),
             'in+': ('lx', -1.25, 0.5),
             'in-': ('lx', -1.25, -0.5),
             'vdd': ('t', 0, 0.5),
             'vdd2': ('t', -0.45, 0.755),
             'vss2': ('b', -0.45, -0.755),
             'vss': ('b', 0, -0.5),
             'ref': ('b', 0.45, -0.245),
             'r+': ('l', -0.85, 0.25),
             'r-': ('l', -0.85, -0.25)}

    npins = {'out': ('rx', 1.25, 0.0),
             'in-': ('lx', -1.25, 0.5),
             'in+': ('lx', -1.25, -0.5),
             'vdd': ('t', 0, 0.5),
             'vdd2': ('t', -0.45, 0.755),
             'vss2': ('b', -0.45, -0.755),
             'vss': ('b', 0, -0.5),
             'ref': ('b', 0.45, -0.245),
             'r-': ('l', -0.85, 0.25),
             'r+': ('l', -0.85, -0.25)}

    extra_fields = {'mirror': 'Mirror', 'invert': 'Invert'}

    def assign_positions(self, x1, y1, x2, y2) -> array:
        """Assign node positions based on cursor positions.

        x1, y1 defines the positive input node
        x2, y2 defines the negative input node"""

        r = sqrt((x2 - x1)**2 + (y2 - y1)**2)

        # TODO: handle rotation

        xo = (x2 + x1) / 2 + r * 5 / 2
        yo = (y2 + y1) / 2

        positions = array(((xo, yo),
                           (nan, nan),
                           (x1, y1),
                           (x1, y2)))
        return positions

    @property
    def midpoint(self):

        pos = (self.nodes[2].pos + self.nodes[3].pos) * 0.5

        return (self.nodes[0].pos + pos) * 0.5

    @property
    def length(self) -> float:

        pos = (self.nodes[2].pos + self.nodes[3].pos) * 0.5

        diff = (pos - self.nodes[0].pos) * 0.5
        return diff.norm()

    def attr_dir_string(self, x1, y1, x2, y2, step=1):

        # TODO: Handle rotation
        dy = abs(y2 - y1)
        size = dy * 5 / 4

        attr = 'right=%s' % size
        return attr

    def draw(self, editor, sketcher, **kwargs):

        x1, y1 = self.nodes[2].pos.x, self.nodes[2].pos.y
        x2, y2 = self.nodes[3].pos.x, self.nodes[3].pos.y

        xc = (x1 + x2) / 2
        yc = (y1 + y2) / 2

        dy = abs(self.nodes[3].y - self.nodes[2].y)
        size = dy * 5 / 4

        kwargs = self.make_kwargs(editor, **kwargs)

        sketcher.sketch(self.sketch, offset=(xc, yc), angle=0, scale=size / 2.5,
                        **kwargs)

    def netitem_nodes(self, node_names):

        parts = []
        for node_name in node_names[0:2]:
            parts.append(node_name)
        parts.append('opamp')
        for node_name in node_names[2:]:
            parts.append(node_name)
        return parts

    def is_within_bbox(self, x, y):

        x0, y0 = self.nodes[0].pos.x, self.nodes[0].pos.y
        x1, y1 = self.nodes[2].pos.x, self.nodes[2].pos.y
        x2, y2 = self.nodes[3].pos.x, self.nodes[3].pos.y

        # TODO, adjust for actual triangle

        return point_in_triangle(x, y, x0, y0, x1, y1, x2, y2)

    @property
    def node1(self):

        return self.nodes[2]

    @property
    def node2(self):

        return self.nodes[3]

    def choose_node_name(self, m, nodes):

        # Handle the special ground node.
        if m == 1:
            return '0'
        return super().choose_node_name(m, nodes)



from tkinter import Tk
from .labelentries import LabelEntry, LabelEntries


class ExprAttributesDialog:

    def __init__(self, expr, ui, title='Expression attributes'):

        self.expr = expr
        self.ui = ui

        self.window = Tk()
        self.window.title(title)

        entries = [LabelEntry('units', 'Units', expr.units),
                   LabelEntry('domain', 'Domain', expr.domain),
                   LabelEntry('quantity', 'Quantity', expr.quantity),
                   LabelEntry('causal', 'Causal', expr.is_causal, command=self.causal)]

        self.labelentries = LabelEntries(self.window, ui, entries)

    def causal(self):

        self.expr.is_causal = self.labelentries.get('causal')



from tkinter import Tk, StringVar, Label, OptionMenu, Entry, Button
from ..components import Capacitor, Inductor


class CptDialog:

    def __init__(self, cpt, update=None):

        self.cpt = cpt
        self.update = update

        self.window = Tk()
        self.window.title(cpt.name)

        row = 0

        self.kind_var = None
        if cpt.kind is not None:
            self.kind_var = StringVar(self.window)
            self.kind_var.set(cpt.kind)

            kind_label = Label(self.window, text='Kind: ')
            kind_option = OptionMenu(self.window, self.kind_var,
                                     *cpt.kinds.keys())

            kind_label.grid(row=row)
            kind_option.grid(row=row, column=1)
            row += 1

        self.name_var = StringVar(self.window)
        self.name_var.set(cpt.name)

        name_label = Label(self.window, text='Name: ')
        name_entry = Entry(self.window, textvariable=self.name_var,
                           command=self.on_update)

        name_label.grid(row=row)
        name_entry.grid(row=row, column=1)
        row += 1

        self.value_var = StringVar(self.window)
        value = cpt.value
        if value is None:
            value = cpt.name
        self.value_var.set(value)

        value_label = Label(self.window, text='Value: ')
        value_entry = Entry(self.window, textvariable=self.value_var,
                            command=self.on_update)

        value_label.grid(row=row)
        value_entry.grid(row=row, column=1)
        row += 1

        self.initial_value_var = None
        if isinstance(cpt, (Capacitor, Inductor)):

            ivlabel = 'v0'
            if isinstance(cpt, Inductor):
                ivlabel = 'i0'

            self.initial_value_var = StringVar(self.window)

            initial_value_label = Label(self.window, text=ivlabel + ': ')
            initial_value_entry = Entry(
                self.window, textvariable=self.initial_value_var,
                command=self.on_update)
            initial_value_label.grid(row=row)
            initial_value_entry.grid(row=row, column=1)
            row += 1

        button = Button(self.window, text="OK", command=self.on_ok)
        button.grid(row=row)

    def on_update(self, arg=None):

        if self.kind_var is not None:
            kind = self.kind_var.get()
            if kind == '':
                kind = None
            self.cpt.kind = kind

        self.cpt.name = self.name_var.get()

        value = self.value_var.get()
        if value == '':
            value = None
        self.cpt.value = value

        if self.initial_value_var is not None:
            value = self.initial_value_var.get()
            if value == '':
                value = None
            self.cpt.initial_value = value

        if self.update:
            self.update(self.cpt)

    def on_ok(self):

        self.on_update()

        self.window.destroy()



from .component import Component


class BipoleComponent(Component):

    can_stretch = True

    @property
    def sketch_net(self):

        s = self.type + ' 1 2; right'
        if self.symbol_kind != '':
            s += ', kind=' + self.symbol_kind
        if self.style != '':
            s += ', style=' + self.style
        return s



from .admittance import Admittance
from .bjt import BJT
from .capacitor import Capacitor
from .connection import Connection
from .cpe import CPE
from .current_source import CurrentSource
from .diode import Diode
from .ferritebead import FerriteBead
from .impedance import Impedance
from .inductor import Inductor
from .jfet import JFET
from .mosfet import MOSFET
from .opamp import Opamp
from .opencircuit import OpenCircuit
from .port import Port
from .resistor import Resistor
from .voltage_source import VoltageSource
from .wire import Wire
from .vcvs import VCVS
from .vccs import VCCS
from .ccvs import CCVS
from .cccs import CCCS

from .sketch import Sketch

# Could use importlib.import_module to programmatically import
# the component classes.


class CptMaker:

    cpts = {
        'C': Capacitor,
        'CPE': CPE,
        'D': Diode,
        'E': VCVS,
        'opamp': Opamp,
        'F': CCCS,
        'FB': FerriteBead,
        'G': VCCS,
        'H': CCVS,
        'I': CurrentSource,
        'J': JFET,
        'L': Inductor,
        'M': MOSFET,
        'O': OpenCircuit,
        'P': Port,
        'Q': BJT,
        'R': Resistor,
        'NR': Resistor,         # Noise free resistor
        'V': VoltageSource,
        'W': Wire,
        'X': Connection,
        'Y': Admittance,
        'Z': Impedance
    }

    def __init__(self):

        self.sketches = {}

    def _make_cpt(self, cpt_type, kind='', style='', name=None,
                  nodes=None, opts=None):

        if cpt_type == 'W' and kind != '':
            cls = Connection
        elif cpt_type == 'E' and kind == 'opamp':
            cls = Opamp
        else:
            cls = self.cpts[cpt_type]

        cpt = cls(kind=kind, style=style,
                  name=name, nodes=nodes, opts=opts)
        return cpt

    def _add_sketch(self, cpt):

        sketch_key = cpt.sketch_key

        try:
            sketch = self.sketches[sketch_key]
        except KeyError:
            sketch = Sketch.load(cpt.sketch_key)
            if sketch is None:
                raise FileNotFoundError(
                    'Could not find data file for ' + cpt.sketch_key)
            self.sketches[sketch_key] = sketch

        # TODO: remove duck typing
        cpt.sketch = sketch

    def __call__(self, cpt_type, kind='', style='', name=None,
                 nodes=None, opts=None, add_sketch=True):

        cpt = self._make_cpt(cpt_type, kind, style, name, nodes, opts)

        if add_sketch:
            self._add_sketch(cpt)

        return cpt


cpt_maker = CptMaker()


def cpt_make_from_cpt(cpt):

    ctype = cpt.type

    # Convert wire with implicit connection to a connection component.
    if ctype == 'W':
        for kind in Connection.kinds:
            if kind[1:] in cpt.opts:
                ctype = 'X'
                break

    return cpt_maker(ctype, kind=cpt._kind, name=cpt.name,
                     nodes=cpt.nodes, opts=cpt.opts)


def cpt_make_from_type(cpt_type, cpt_name='', kind='', style='',
                       add_sketch=True):

    return cpt_maker(cpt_type, name=cpt_name, kind=kind, style=style,
                     add_sketch=add_sketch)


def cpt_remake(cpt):

    return cpt_maker._add_sketch(cpt)


def cpt_sketch_make(cpt):

    # Could make method of cpt.
    sketch = Sketch.create(cpt.sketch_key, cpt.sketch_net)
    return sketch



from .transistor import Transistor


class JFET(Transistor):

    type = "J"
    angle_offset = 90
    default_kind = 'njf-'

    kinds = {'njf-': 'NJFET',
             'pjf-': 'PJFET'}

    # TODO: add gate offset



from tkinter import Tk, Button
from .labelentries import LabelEntry, LabelEntries


class TwoportDialog:

    def __init__(self, ui, cpt, kind):

        self.ui = ui
        self.kind = kind
        self.window = Tk()
        self.window.title('Twoport')

        entries = []

        elements = ui.model.circuit.elements

        # TODO, allow V or I
        names = [elt.name for elt in elements.values()
                 if elt.type not in ('W', 'O', 'V', 'I')]

        entries.append(LabelEntry('input', 'Input',
                                  names[0], names))

        entries.append(LabelEntry('output', 'Output',
                                  names[-1], names))

        self.labelentries = LabelEntries(self.window, ui, entries)

        button = Button(self.window, text="Create", command=self.on_create)
        button.grid(row=self.labelentries.row)

        self.window.minsize(200, 50)

    def on_create(self):

        input_cpt = self.labelentries.get('input')
        output_cpt = self.labelentries.get('output')

        # Remove independent sources
        cct = self.ui.model.circuit
        cct = cct.copy()
        values = list(cct.elements.values())
        for cpt in values:
            if cpt.is_independent_source:
                cct.remove(cpt.name)

        # Warn that might be slow
        A = cct.twoport(input_cpt, output_cpt, model=self.kind)

        self.ui.show_twoport_select_dialog(A, self.kind)



"""
Root-level lcapy-gui objects.
These can be imported directly from lcapy-gui.
"""

from .ui.tk.lcapytk import LcapyTk
import sys

if sys.version_info < (3, 8):
    import pkg_resources
    __version__ = pkg_resources.require('lcapygui')[0].version
else:
    from importlib.metadata import version
    __version__ = version('lcapygui')


if sys.version_info < (3, 9):
    import importlib_resources
else:
    import importlib.resources as importlib_resources

pkg = importlib_resources.files('lcapygui')
__datadir__ = pkg / 'data'

__libdir__ = __datadir__ / 'lib'



from tkinter import Tk, Button
from numpy import linspace
from .labelentries import LabelEntry, LabelEntries


class PlotPropertiesDialog:

    def __init__(self, expr, ui):

        self.expr = expr
        self.ui = ui

        self.window = Tk()
        self.window.title('Plot properties')

        entries = [LabelEntry('min', 'Min', 0.0),
                   LabelEntry('max', 'Max', 1.0),
                   LabelEntry('points', 'Points', 200)]

        self.symbols = []
        for key in expr.symbols:
            # Ignore domain variable
            if key != expr.var.name:
                entries.append(LabelEntry(key, key, 0.0))
                self.symbols.append(key)

        plots = {'Plot': 'plot', 'Bode': 'bode_plot',
                 'Pole-zero': 'pole_zero_plot', 'Nichols': 'nichols_plot',
                 'Nyquist': 'nyquist_plot'}

        kinds = []
        for key, val in plots.items():
            if hasattr(expr, val):
                kinds.append(key)

        entries.append(LabelEntry('kind', 'Plot type', 'Plot', kinds))
        self.labelentries = LabelEntries(self.window, ui, entries)

        button = Button(self.window, text="Plot", command=self.on_update)
        button.grid(row=self.labelentries.row)

    def on_update(self):

        points = linspace(self.labelentries.get('min'),
                          self.labelentries.get('max'),
                          self.labelentries.get('points'))

        defs = {}
        for key in self.symbols:
            val = self.labelentries.get_text(key)
            if val == '':
                self.ui.show_error_dialog('Undefined symbol ' + key)
                return
            val = self.labelentries.get(key)
            defs[key] = val

        expr = self.expr.subs(defs)

        kind = self.labelentries.get('kind')
        if kind == 'Plot':
            im = expr.plot(points)
        elif kind == 'Bode':
            im = expr.bode_plot(points)
        elif kind == 'Pole-zero':
            im = expr.pole_zero_plot()
        elif kind == 'Nichols':
            im = expr.nichols_plot(points)
        elif kind == 'Nyquist':
            im = expr.nyquist_plot(points)
        else:
            raise RuntimeError('Unexpected case')

        if isinstance(im, tuple):
            im = im[0]

        im.figure.show()



from .component import Component
from numpy import array, sqrt, dot, degrees
from numpy.linalg import norm
from math import atan2


class Transistor(Component):

    can_stretch = True
    label_offset = 0.6
    has_value = False
    # extra_fields = {'mirror': 'Mirror', 'invert': 'Invert'}

    def assign_positions(self, x1, y1, x2, y2) -> array:
        """Assign node positions based on cursor positions.

        x1, y1 defines the positive input node
        x2, y2 defines the negative input node"""

        dx = x1 - x2
        dy = y1 - y2
        r = sqrt(dx**2 + dy**2)
        R = array(((dx, -dy), (dy, dx))) / r

        # Midpoint
        ym = (y1 + y2) / 2
        xm = (x1 + x2) / 2

        # TODO: The offset might change with transistor types.
        # Note, the transistor base/gate is down.
        xg, yg = dot(R, (0, 2))
        xg += xm
        yg += ym

        # D G S
        positions = array(((x1, y1),
                           (xg, yg),
                           (x2, y2)))
        return positions

    @property
    def node1(self):

        return self.nodes[0]

    @property
    def node2(self):

        return self.nodes[2]

    @property
    def sketch_net(self):

        # With up, drain is down.
        s = self.type + ' 1 2 3 ' + self.cpt_kind + '; up'
        if self.symbol_kind != '':
            s += ', kind=' + self.symbol_kind
        return s



from tkinter import Tk, Button
from .labelentries import LabelEntry, LabelEntries


class CptPropertiesDialog:

    def __init__(self, ui, cpt, update=None, title=''):

        self.cpt = cpt
        self.gcpt = cpt.gcpt
        self.update = update
        self.ui = ui

        self.window = Tk()
        self.window.title(title)

        entries = []
        if self.gcpt.kinds != {}:
            kind_name = self.gcpt.kinds[self.gcpt.kind]
            entries.append(LabelEntry(
                'kind', 'Kind', kind_name, list(self.gcpt.kinds.values()),
                command=self.on_update))

        if self.gcpt.styles != {}:
            style_name = self.gcpt.styles[self.gcpt.style]
            entries.append(LabelEntry(
                'style', 'Style', style_name, list(self.gcpt.styles.values()),
                command=self.on_update))

        entries.append(LabelEntry('name', 'Name', self.cpt.name,
                                  command=self.on_update))

        for m, arg in enumerate(self.gcpt.args):
            if arg == 'Control':
                continue
            entries.append(LabelEntry(arg, arg, self.cpt.args[m],
                                      command=self.on_update))

        if cpt.is_capacitor:
            v0 = str(self.cpt.cpt.v0) if self.cpt.cpt.has_ic else ''
            entries.append(LabelEntry('v0', 'v0', v0,
                                      command=self.on_update))
        elif cpt.is_inductor:
            i0 = str(self.cpt.cpt.i0) if self.cpt.cpt.has_ic else ''
            entries.append(LabelEntry('i0', 'i0', i0,
                                      command=self.on_update))
        elif cpt.is_dependent_source and self.gcpt.type != 'Eopamp':
            names = ui.model.possible_control_names()
            entries.append(LabelEntry('control', 'Control',
                                      self.gcpt.control, names,
                                      command=self.on_update))

        for k, v in self.gcpt.fields.items():
            entries.append(LabelEntry(k, v, getattr(self.gcpt, k),
                                      command=self.on_update))

        for k, v in self.gcpt.extra_fields.items():
            entries.append(LabelEntry(k, v, getattr(self.gcpt, k),
                                      command=self.on_update))

        self.labelentries = LabelEntries(self.window, ui, entries)

        button = Button(self.window, text="OK", command=self.on_ok)
        button.grid(row=self.labelentries.row)

    def on_update(self, arg=None):

        if self.gcpt.kinds != {}:
            kind = self.gcpt.inv_kinds[self.labelentries.get('kind')]
            if self.gcpt.kind != kind:
                self.gcpt.kind = kind
                # Need a new cpt

        if self.gcpt.styles != {}:
            self.gcpt.style = self.gcpt.inv_styles[self.labelentries.get(
                'style')]

        name = self.labelentries.get('name')
        if name.startswith(self.gcpt.name[0]):
            self.gcpt.name = self.labelentries.get('name')
        else:
            self.ui.show_error_dialog('Cannot change component type')

        for m, arg in enumerate(self.gcpt.args):
            if arg == 'Control':
                continue
            value = self.labelentries.get(arg)
            if value == '':
                value = self.gcpt.name
            self.cpt.args[m] = value

        try:
            if self.cpt.is_capacitor:
                v0 = self.labelentries.get('v0')
                self.cpt.args[-1] = v0
                if v0 == '':
                    v0 = None
                else:
                    try:
                        v0 = float(v0)
                    except ValueError:
                        # Symbolic
                        pass

                cpt = self.cpt.cpt
                # Create new oneport (this should be improved).
                self.cpt._cpt = cpt.__class__(cpt.C, v0)
            elif self.cpt.is_inductor:
                i0 = self.labelentries.get('i0')
                self.cpt.args[-1] = i0
                if i0 == '':
                    i0 = None
                else:
                    try:
                        i0 = float(i0)
                    except ValueError:
                        # Symbolic
                        pass
                cpt = self.cpt.cpt
                # Create new oneport (this should be improved).
                self.cpt._cpt = cpt.__class__(cpt.L, i0)
        except KeyError:
            pass

        try:
            self.gcpt.control = self.labelentries.get('control')
        except KeyError:
            pass

        for k, v in self.gcpt.fields.items():
            setattr(self.gcpt, k, self.labelentries.get(k))

        for k, v in self.gcpt.extra_fields.items():
            setattr(self.gcpt, k, self.labelentries.get(k))

        if self.update:
            self.update(self.cpt)

    def on_ok(self):

        self.on_update()

        self.window.destroy()



class Annotations(list):

    def __init__(self):

        super(Annotations, self).__init__(self)

    def add(self, annotation):

        self.append(annotation)

    def remove(self):

        while self != []:
            self.pop().remove()



from ..annotation import Annotation
from ..annotations import Annotations
from .preferences import Preferences
from ..components.opamp import Opamp
from ..components.cpt_maker import cpt_make_from_cpt, cpt_make_from_type, cpt_remake

from copy import copy
from math import atan2, degrees, sqrt
from numpy import nan, isnan
from lcapy import Circuit, expr
from lcapy.mnacpts import Cpt
from lcapy.nodes import parse_nodes
from lcapy.schemmisc import Pos
from lcapy.opts import Opts


class UIModelBase:

    STEP = 2
    SNAP = 1
    SCALE = 0.25

    # Short-cut key, name, type, kind
    component_map = {
        'c': ('c', 'Capacitor', 'C', ''),
        'd': ('d', 'Diode', 'D', ''),
        'fb': ('', 'Ferrite bead', 'FB', ''),
        'i': ('i', 'Current source', 'I', ''),
        'j': ('j', 'JFET', 'J', ''),
        'l': ('l', 'Inductor', 'L', ''),
        'm': ('m', 'MOSFET', 'M', ''),
        'q': ('q', 'BJT', 'Q', ''),
        'r': ('r', 'Resistor', 'R', ''),
        'nr': ('', 'Noiseless resistor', 'R', ''),
        'v': ('v', 'Voltage source', 'V', ''),
        'w': ('w', 'Wire', 'W', ''),
        'e': ('e', 'VCVS', 'E', ''),
        'f': ('f', 'CCCS', 'F', ''),
        'g': ('g', 'VCCS', 'G', ''),
        'h': ('h', 'CCVS', 'H', ''),
        'o': ('o', 'Open circuit', 'O', ''),
        'opamp': ('', 'Opamp', 'opamp', ''),
        'p': ('p', 'Port', 'P', ''),
        'y': ('y', 'Admittance', 'Y', ''),
        'z': ('z', 'Impedance', 'Z', ''),
        'cpe': ('', 'CPE', 'CPE', ''),
    }

    connection_map = {
        '0': ('0', '0V', 'W', ''),
        '0V': ('', '0V', 'W', '0V'),
        'ground': ('', 'Ground', 'W', ''),
        'sground': ('', 'Signal ground', 'W', 'sground'),
        'rground': ('', 'Rail ground', 'W', 'rground'),
        'cground': ('', 'Chassis ground', 'W', 'cground'),
        'vdd': ('', 'VDD', 'W', 'vdd'),
        'vss': ('', 'VSS', 'W', 'vss'),
        'vcc': ('', 'VCC', 'W', 'vcc'),
        'vee': ('', 'VEE', 'W', 'vee'),
        'input': ('', 'Input', 'W', 'input'),
        'output': ('', 'Output', 'W', 'output'),
        'bidir': ('', 'Bidirectional', 'W', 'bidir')
    }

    def __init__(self, ui):

        self.circuit = Circuit()
        self.ui = ui
        self._analysis_circuit = None
        self.pathname = ''
        self.voltage_annotations = Annotations()
        self.selected = None
        self.last_expr = None
        self.preferences = Preferences()
        self.dirty = False
        self.history = []
        self.clipboard = None

    @property
    def analysis_circuit(self):
        """This like circuit but it has an added ground node if one does
        not exist.

        """

        if self._analysis_circuit is not None:
            return self._analysis_circuit

        if self.circuit.elements == {}:
            self.exception('No circuit defined')
            return None

        self._analysis_circuit = self.circuit.copy()

        if self.ground_node is None:
            # Add dummy ground node to first node
            net = 'W %s 0\n' % list(self.circuit.nodes)[0]
            self.analysis_circuit.add(net)

        try:
            self._analysis_circuit[0]
        except (AttributeError, ValueError, RuntimeError) as e:
            self.exception(e)
            return None

        return self._analysis_circuit

    def bounding_box(self):

        if len(self.circuit.nodes) == 0:
            return None

        xmin = 1000
        xmax = 0
        ymin = 1000
        ymax = 0
        for node in self.circuit.nodes.values():
            if node.x < xmin:
                xmin = node.x
            if node.x > xmax:
                xmax = node.x
            if node.y < ymin:
                ymin = node.y
            if node.y > ymax:
                ymax = node.y
        return xmin, ymin, xmax, ymax

    def choose_cpt_name(self, cpt_type):

        if cpt_type == 'opamp':
            cpt_type = 'E'

        num = 1
        while True:
            name = cpt_type + str(num)
            if name not in self.circuit.elements:
                return name
            num += 1

    def con_create(self, con_key, x1, y1, x2, y2):
        """Create a new connection."""

        try:
            cpt_type = self.connection_map[con_key][2]
        except KeyError:
            return None

        if cpt_type == '':
            return None

        return self.thing_create(cpt_type, x1, y1, x2, y2, kind='-' + con_key)

    def copy(self, cpt):

        self.clipboard = cpt

    @property
    def cpt_selected(self):

        return isinstance(self.selected, Cpt)

    def cpt_create(self, cpt_key, x1, y1, x2, y2):
        """Create a new component."""

        s = sqrt((x1 - x2)**2 + (y1 - y2)**2)
        if s < 0.2:
            self.exception('Nodes too close to create component')
            return

        try:
            cpt_type = self.component_map[cpt_key][2]
        except KeyError:
            return None

        if cpt_type == '':
            return None

        return self.thing_create(cpt_type, x1, y1, x2, y2)

    def cpt_delete(self, cpt):

        self.select(None)

        redraw = True
        try:
            # This should also delete the annotations.
            cpt.undraw()
            redraw = False
        except AttributeError:
            pass

        self.circuit.remove(cpt.name)
        self.invalidate()

        if redraw:
            self.ui.clear()
            self.redraw()

    def cpt_draw(self, cpt):

        gcpt = cpt.gcpt
        if gcpt is None:
            return

        gcpt.draw(self, self.ui.sketcher)

        label_cpts = self.preferences.label_cpts

        if gcpt.type in ('A', 'O', 'W'):
            label_cpts = 'none'

        name = cpt.name

        try:
            if cpt.type in ('F', 'H'):
                value = cpt.args[1]
            elif cpt.type in ('P', ):
                value = None
            else:
                value = cpt.args[0]
        except IndexError:
            value = None

        if value is None:
            value = ''
            value_latex = ''
        else:
            value_latex = '$' + expr(value).latex() + '$'

        label = ''
        if label_cpts == 'name+value':
            if name != value and gcpt.has_value:
                label = name + '=' + value_latex
            else:
                label = name
        elif label_cpts == 'value':
            if value != '':
                label = value_latex
        elif label_cpts == 'name':
            label = name
        elif label_cpts == 'none':
            label = ''
        else:
            raise RuntimeError('Unhandled label_cpts=' + label_cpts)

        if label != '':
            ann = Annotation(self.ui, gcpt.label_position.x,
                             gcpt.label_position.y, label)
            ann.draw(fontsize=18)
            gcpt.annotations.append(ann)

        draw_nodes = self.preferences.draw_nodes
        if draw_nodes != 'none':
            for node in gcpt.drawn_nodes:
                if node.port:
                    self.node_draw(node)
                    continue

                if draw_nodes == 'connections' and node.count < 3:
                    continue
                if draw_nodes == 'primary' and not node.is_primary:
                    continue
                self.node_draw(node)

        label_nodes = self.preferences.label_nodes
        if label_nodes != 'none':
            for node in gcpt.labelled_nodes:
                if node.name[0] == '_':
                    continue

                if label_nodes == 'alpha' and not node.name[0].isalpha():
                    continue

                x, y = node.pos.x, node.pos.y
                # Should be x -= 0.1 but need to right justify.
                x += 0.1
                y += 0.1
                ann = Annotation(self.ui, x, y, node.name)
                ann.draw(fontsize=18)
                gcpt.annotations.append(ann)

    def cpt_find(self, node_name1, node_name2):

        fcpt = None
        for cpt in self.circuit:
            if (cpt.nodes[0].name == node_name1 and cpt.nodes[1].name == node_name2):
                fcpt = cpt
                break
        if fcpt is None:
            self.exception(
                'Cannot find a component with nodes %s and %s' % (node_name1, node_name2))
        return fcpt

    def cpt_remake(self, cpt):

        gcpt = cpt.gcpt

        if cpt.is_dependent_source and gcpt.type != 'Eopamp':
            try:
                newcpt = cpt._change_control(gcpt.control)
            except Exception:
                self.exception('Control component %s for %s deleted' %
                               (gcpt.control, cpt.name))
                return
        elif gcpt.cpt_kind == cpt._kind:
            newcpt = cpt
        elif gcpt.type != 'Eopamp':
            try:
                newcpt = cpt._change_kind(gcpt.cpt_kind)
            except Exception:
                self.exception('Cannot change kind for %s' % cpt.name)
                return
        else:
            newcpt = cpt

        if gcpt.name != cpt.name:
            try:
                newcpt = newcpt._change_name(gcpt.name)
            except Exception:
                self.exception('Cannot change name for %s' % cpt.name)
                return

        if gcpt.mirror ^ ('mirror' in newcpt.opts):
            # TODO, add mirror method...
            if gcpt.type == 'Eopamp':
                newcpt.nodes[2], newcpt.nodes[3] = newcpt.nodes[3], newcpt.nodes[2]
            elif gcpt.type in ('J', 'M', 'Q'):
                newcpt.nodes[2], newcpt.nodes[0] = newcpt.nodes[0], newcpt.nodes[2]
            else:
                print('Trying to change mirror for ' + str(newcpt))

        newcpt.opts.clear()
        newcpt.opts.add(gcpt.attr_string_update(self.STEP))

        newcpt.gcpt = gcpt
        cpt_remake(newcpt.gcpt)

    def cut(self, cpt):

        self.delete(cpt)
        self.clipboard = cpt

    def delete(self, cpt):

        self.cpt_delete(cpt)
        self.history.append((cpt, 'D'))

    def draw(self, cpt, **kwargs):

        if cpt is None:
            return
        cpt.draw(**kwargs)

    def export(self, pathname):

        cct = self.circuit
        cct.draw(pathname)

    def invalidate(self):

        self._analysis_circuit = None

    def load(self, pathname):

        from lcapy import Circuit

        self.pathname = pathname

        with open(pathname) as f:
            line = f.readline()
            if line.startswith(r'\begin{tikz'):
                self.ui.show_error_dialog('Cannot load Circuitikz macro file')
                return

        try:
            circuit = Circuit(pathname)
        except Exception as e:
            self.exception(e)
            return

        return self.load_from_circuit(circuit)

    def load_from_circuit(self, circuit):

        self.circuit = circuit
        positions = None
        for cpt in self.circuit.elements.values():
            if cpt.type == 'XX' and 'nodes' in cpt.opts:
                positions = parse_nodes(cpt.opts['nodes'])
                break

        if positions is not None:
            for k, v in self.circuit.nodes.items():
                try:
                    v.pos = positions[k]
                except KeyError:
                    v.pos = None

        else:

            # Node positions not defined.

            sch = self.circuit.sch

            try:
                # This will fail if have detached components.
                calculated = sch._positions_calculate()
            except (AttributeError, ValueError, RuntimeError) as e:
                self.exception(e)
                return

            width, height = sch.width * self.STEP, sch.height * self.STEP

            # Centre the schematic.
            xsize = self.ui.canvas.drawing.xsize
            ysize = self.ui.canvas.drawing.ysize
            offsetx, offsety = self.snap((xsize - width) / 2,
                                         (ysize - height) / 2)
            for node in sch.nodes.values():
                node.pos.x += offsetx
                node.pos.y += offsety
                # May have split nodes...
                if node.name in circuit.nodes:
                    circuit.nodes[node.name].pos = node.pos

        self.remove_directives()

        for cpt in self.circuit.elements.values():
            if cpt.type == 'XX':
                cpt.gcpt = None
                continue
            try:
                gcpt = cpt_make_from_cpt(cpt)
            except Exception as e:
                gcpt = None
                self.exception(e)

            cpt.gcpt = gcpt

        self.invalidate()
        self.redraw()

    def move(self, xshift, yshift):
        # TODO
        pass

    def paste(self, x1, y1, x2, y2):

        if self.clipboard is None:
            return

        return self.thing_create(self.clipboard.type, x1, y1, x2, y2)

    def possible_control_names(self):

        cpts = self.circuit.elements.values()
        names = [c.name for c in cpts if c.name[0] != 'W']
        return names

    def remove_directives(self):

        elt_list = list(self.circuit.elements.values())
        if elt_list == []:
            return

        cpt = elt_list[-1]
        if cpt.type == 'XX':
            # TODO: make more robust
            # This tries to remove the schematic attributes.
            # Perhaps parse this and set preferences but this
            # might be confusing.
            self.circuit.remove(cpt.name)
            cpt = elt_list[0]

        if cpt.type == 'XX' and cpt._string.startswith('# Created by lcapy'):
            self.circuit.remove(cpt.name)

        if len(elt_list) > 1:
            cpt = elt_list[1]
            if cpt.type == 'XX' and cpt._string.startswith('; nodes='):
                self.circuit.remove(cpt.name)

    def rotate(self, angle):
        # TODO
        pass

    def save(self, pathname):

        s = self.schematic()

        with open(pathname, 'w') as fhandle:
            fhandle.write(s)
        self.dirty = False

    def schematic(self):

        s = '# Created by ' + self.ui.NAME + ' V' + self.ui.version + '\n'

        # Define node positions
        foo = [str(node) for node in self.circuit.nodes.values()
               if not isnan(node.pos.x)]

        s += '; nodes={' + ', '.join(foo) + '}' + '\n'

        for cpt in self.circuit.elements.values():
            s += str(cpt) + '\n'

        # FIXME, remove other preference string
        # Note, need a newline so string treated as a netlist string
        s += '; ' + self.preferences.schematic_preferences() + '\n'
        return s

    def thing_create(self, cpt_type, x1, y1, x2, y2, kind=''):

        cpt_name = self.choose_cpt_name(cpt_type)
        gcpt = cpt_make_from_type(cpt_type, cpt_name, kind=kind)
        if gcpt is None:
            return

        all_node_names = list(self.circuit.nodes)
        node_names = []
        positions = gcpt.assign_positions(x1, y1, x2, y2)

        for m, position in enumerate(positions):
            if position is None:
                continue

            node = self.circuit.nodes.by_position(position)
            if node is None:
                node_name = gcpt.choose_node_name(m, all_node_names)
                all_node_names.append(node_name)
            else:
                node_name = node.name
            node_names.append(node_name)

        netitem = gcpt.netitem(node_names, x1, y1, x2, y2, self.STEP)

        if self.ui.debug:
            print('Adding ' + netitem)

        cct = self.circuit.add(netitem)
        self.invalidate()
        cpt = cct[cpt_name]

        for m, position in enumerate(positions):
            cpt.nodes[m].pos = Pos(position)

        attr_string = netitem.split(';', 1)[1]
        gcpt.update(nodes=cpt.nodes, opts=Opts(attr_string))

        # Duck type
        cpt.gcpt = gcpt

        self.cpt_draw(cpt)

        self.history.append((cpt, 'A'))

        return cpt

    def inspect_admittance(self, cpt):

        try:
            self.last_expr = self.analysis_circuit[cpt.name].Y
            self.ui.show_expr_dialog(self.last_expr,
                                     '%s admittance' % cpt.name)
        except (AttributeError, ValueError, RuntimeError) as e:
            self.exception(e)

    def inspect_current(self, cpt):

        # TODO: FIXME for wire current
        try:
            self.last_expr = self.analysis_circuit[cpt.name].i
            self.ui.show_expr_dialog(self.last_expr,
                                     '%s current' % cpt.name)
        except (AttributeError, ValueError, RuntimeError) as e:
            self.exception(e)

    def inspect_impedance(self, cpt):

        try:
            self.last_expr = self.analysis_circuit[cpt.name].Z
            self.ui.show_expr_dialog(self.last_expr,
                                     '%s impe' % cpt.name)
        except (AttributeError, ValueError, RuntimeError) as e:
            self.exception(e)

    def inspect_norton_admittance(self, cpt):

        try:
            self.last_expr = self.analysis_circuit[cpt.name].dpY
            self.ui.show_expr_dialog(self.last_expr,
                                     '%s Norton admittance' % cpt.name)
        except (AttributeError, ValueError, RuntimeError) as e:
            self.exception(e)

    def inspect_thevenin_impedance(self, cpt):

        try:
            self.last_expr = self.analysis_circuit[cpt.name].dpZ
            self.ui.show_expr_dialog(self.last_expr,
                                     '%s Thevenin impedance' % cpt.name)
        except (AttributeError, ValueError, RuntimeError) as e:
            self.exception(e)

    def inspect_voltage(self, cpt):

        try:
            self.last_expr = self.analysis_circuit[cpt.name].v
            self.ui.show_expr_dialog(self.last_expr,
                                     '%s potential difference' % cpt.name)
        except (AttributeError, ValueError, RuntimeError) as e:
            self.exception(e)

    def show_node_voltage(self, node):

        try:
            self.last_expr = self.analysis_circuit[node.name].v
            self.ui.show_expr_dialog(self.last_expr,
                                     'Node %s potential' % node.name)
        except (AttributeError, ValueError, RuntimeError) as e:
            self.exception(e)

    def select(self, thing):

        self.selected = thing

    def snap(self, x, y):

        snap = self.SNAP
        x = (x + 0.5 * snap) // snap * snap
        y = (y + 0.5 * snap) // snap * snap
        return x, y

    def unselect(self):
        pass

    def view(self):

        cct = Circuit(self.schematic())
        cct.draw()

    def voltage_annotate(self, cpt):

        ann1 = Annotation(self.ui, *cpt.nodes[0].pos, '+')
        ann2 = Annotation(self.ui, *cpt.nodes[1].pos, '-')

        self.voltage_annotations.add(ann1)
        self.voltage_annotations.add(ann2)
        ann1.draw(color='red', fontsize=40)
        ann2.draw(color='blue', fontsize=40)

    @property
    def ground_node(self):

        return self.node_find('0')

    def node_draw(self, node):

        if node.pos is None:
            print('Pos unknown for ' + str(node))
            return

        if node.port:
            self.ui.sketcher.stroke_circle(
                node.x, node.y, self.preferences.node_size,
                color=self.preferences.node_color, alpha=1)
        else:
            self.ui.sketcher.stroke_filled_circle(
                node.x, node.y, self.preferences.node_size,
                color=self.preferences.node_color, alpha=1)

    def node_find(self, nodename):

        for node in self.circuit.nodes.values():
            if node.name == nodename:
                return node
        return None

    def redo(self):

        # TODO
        pass

    def redraw(self):

        for cpt in self.circuit.elements.values():
            self.cpt_draw(cpt)

    def undo(self):

        if self.history == []:
            return
        cpt, op = self.history.pop()
        if op == 'D':
            self.circuit.add(str(cpt))

            # Copy node positions
            new_cpt = self.circuit.elements[cpt.name]
            for m, node in enumerate(cpt.nodes):
                new_cpt.nodes[m].pos = node.pos
            new_cpt.gcpt = cpt.gcpt

            self.cpt_draw(cpt)
            self.select(cpt)
        else:
            self.cpt_delete(cpt)
        self.invalidate()



from .bipole import BipoleComponent


class Inductor(BipoleComponent):

    type = 'L'
    default_kind = '-'
    kinds = {'-': '', '-variable': 'Variable', '-choke': 'Choke',
             '-twolineschoke': 'Two lines choke',
             '-sensor': 'Sensor', '-tunable': 'Tunable'}



from tkinter import Tk, Button


class InspectDialog:

    def __init__(self, uimodel, cpt, title=''):

        self.uimodel = uimodel
        self.cpt = cpt

        self.window = Tk()
        self.window.title(title)

        buttons = [('Voltage', self.on_voltage),
                   ('Current', self.on_current),
                   ('Component impedance', self.on_impedance),
                   ('Component admittance', self.on_admittance),
                   ('Thevenin impedance', self.on_thevenin_impedance),
                   ('Norton admittance', self.on_norton_admittance)]
        for b in buttons:
            button = Button(self.window, text=b[0], command=b[1])
            button.pack()

    def on_voltage(self):

        self.uimodel.inspect_voltage(self.cpt)

    def on_current(self):

        self.uimodel.inspect_current(self.cpt)

    def on_impedance(self):

        self.uimodel.inspect_impedance(self.cpt)

    def on_admittance(self):

        self.uimodel.inspect_admittance(self.cpt)

    def on_thevenin_impedance(self):

        self.uimodel.inspect_thevenin_impedance(self.cpt)

    def on_norton_admittance(self):

        self.uimodel.inspect_norton_admittance(self.cpt)



from tkinter import Tk, Button
from numpy import linspace
from .labelentries import LabelEntry, LabelEntries
from .menu import MenuBar, MenuDropdown, MenuItem

# Perhaps iterate over components and annotate values;
# value, v0, phi, omega etc.
# But what about symbols used as component args?
# Better to iterate over undefined symbols in circuit.
# circuit.symbols returns all the symbols in the context;
# this includes domain vars and delta_t etc.


# Workaround for older version of Lcapy
def undefined_symbols(cct):

    from lcapy import expr

    symbols = []
    for k, elt in cct.elements.items():
        cpt = elt.cpt
        for arg in cpt.args:
            symbols += expr(arg).symbols

    return symbols


class EditValuesDialog:

    def __init__(self, ui, title='Component values'):

        self.ui = ui
        self.window = Tk()
        self.window.title(title)
        self.circuit = ui.model.circuit
        try:
            self.symbols = self.circuit.undefined_symbols
        except:
            self.symbols = undefined_symbols(self.circuit)

        menudropdowns = [
            MenuDropdown('File', 0,
                         [MenuItem('Load', self.on_load),
                          ])]

        self.menubar = MenuBar(menudropdowns)
        self.menubar.make(self.window)

        entries = []
        for key in self.symbols:
            entries.append(LabelEntry(key, key, ''))

        self.labelentries = LabelEntries(self.window, ui, entries)

        button = Button(self.window, text="OK",
                        command=self.on_update)
        button.grid(row=self.labelentries.row)

    def on_update(self):

        self.window.destroy()

        defs = {}
        for key in self.symbols:
            val = self.labelentries.get(key)
            if val == '':
                continue
            defs[key] = float(val)

        cct = self.circuit.subs(defs)

        self.ui.model.on_show_new_circuit(cct)

    def on_load(self, arg):

        self.window.destroy()

        pathname = self.ui.open_file_dialog(doc='Definitions', ext='.csv')

        defs = {}
        with open(pathname) as f:
            for line in f.readlines():
                if ',' in line:
                    delim = ','
                else:
                    delim = ' '
                parts = line.split(delim)
                if len(parts) != 2:
                    raise ValueError(
                        'Need comma or space separated definitions')
                name = parts[0].strip()
                value = float(parts[1].strip())
                defs[name] = value

        cct = self.circuit.subs(defs)

        self.ui.model.on_show_new_circuit(cct)



from setuptools import setup, find_packages

# Open readme
with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name='lcapygui',
    packages=find_packages(include=[
        "lcapygui",
        "lcapygui.*"
    ]),
    version="0.9",
    description="A GUI for lcapy",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Michael Hayes, Jordan Hay",
    license="MIT",
    url="https://github.com/mph-/lcapy-gui",
    project_urls={
        "Bug Tracker": "https://github.com/mph-/lcapy-gui",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        "setuptools",
        "importlib",
        "importlib-metadata",
        "importlib-resources",
        "pathlib",
        "lcapy>=1.15",
        "numpy",
        "tk",
        "pillow>=9.4.0",
        "matplotlib",
        "svgpathtools",
        "svgpath2mpl",
        "tkhtmlview"
    ],
    entry_points={
        'console_scripts': [
            'lcapy-tk=lcapygui.scripts.lcapytk:main',
        ],
    },
    include_package_data=True,
    package_data={'': ['data/svg/*.svg', 'data/lib/*.sch']},
    python_requires=">=3.7"  # matched with lcapy
)



from lcapy.system import tmpfilename, LatexRunner, PDFConverter


class LatexImage:

    def __init__(self, s):

        self.s = s

    def image(self):

        tex_filename = tmpfilename('.tex')

        # Need amsmath for operatorname
        template = ('\\documentclass[a4paper]{standalone}\n'
                    '\\usepackage{amsmath}\n'
                    '\\begin{document}\n$%s$\n'
                    '\\end{document}\n')
        content = template % self.s

        open(tex_filename, 'w').write(content)
        pdf_filename = tex_filename.replace('.tex', '.pdf')
        latexrunner = LatexRunner()
        latexrunner.run(tex_filename)

        png_filename = tex_filename.replace('.tex', '.png')
        pdfconverter = PDFConverter()
        pdfconverter.to_png(pdf_filename, png_filename, dpi=300)

        return png_filename



from .transistor import Transistor


class MOSFET(Transistor):

    type = "M"
    angle_offset = 90
    default_kind = 'nmos-nfet'

    kinds = {'nmos-nmos': 'NMOS simple',
             'pmos-pmos': 'PMOS simple',
             'nmos-nmosd': 'NMOS depletion',
             'pmos-pmosd': 'PMOS depletion',
             'nmos-nfet': 'NMOS enhancement',
             'pmos-pfet': 'PMOS enhancement',
             'nmos-nfet-bodydiode': 'NMOS enhancement with body diode',
             'pmos-pfet-bodydiode': 'PMOS enhancement with body diode',
             'nmos-nfetd': 'NMOS depletion',
             'pmos-pfetd': 'PMOS depletion',
             'nmos-nfetd-bodydiode': 'NMOS depletion with body diode',
             'pmos-pfetd-bodydiode': 'PMOS depletion with body diode',
             'nmos-nigfetd': 'NMOS insulated gate depletion',
             'pmos-pigfetd': 'PMOS insulated gate depletion',
             'nmos-nigfetd-bodydiode': 'NMOS insulated gate depletion with bodydiode',
             'pmos-pigfetd-bodydiode': 'PMOS insulated gate depletion with bodydiode',
             'nmos-nigfete': 'NMOS insulated gate enhancement',
             'pmos-pigfete': 'PNMOS insulated gate enhancement',
             'nmos-nigfete-bodydiode': 'NMOS insulated gate enhancement with body diode',
             'pmos-pigfete-bodydiode': 'PMOS insulated gate enhancement with body diode',
             # 'nmos-nigfetebulk': 'nigfetebulk',
             # 'nmos-pigfetebulk': 'pigfetebulk',
             '-hemt': 'HEMT'}

    # TODO: add base offset for nigfetd, pigfetd, nigfete, pigfete,
    # nigfetebulk, pigfetebulk



from .bipole import BipoleComponent


class Resistor(BipoleComponent):

    type = 'R'
    label_offset = 0.4
    default_kind = '-'
    kinds = {'-': '', '-variable': 'Variable',
             '-tunable': 'Tunable'}



from matplotlib.pyplot import subplots, rcParams, show
from matplotlib.backend_tools import ToolBase
from numpy import arange
from sys import exit
from .sketcher import Layer


class Tool(ToolBase):

    def __init__(self, toolmanager, name, func):
        super(Tool, self).__init__(toolmanager, name)
        self.func = func

    def trigger(self, sender, event, data=None):
        self.func()


class DescribeTool(Tool):
    # default_keymap = 'ctrl+?'
    description = 'Describe'


class ExportTool(Tool):
    # default_keymap = 'ctrl+x'
    description = 'Export'


class HelpTool(Tool):
    # default_keymap = 'ctrl+h'
    description = 'Help'


class InspectTool(Tool):
    # default_keymap = 'ctrl+i'
    description = 'Inspect'


class LoadTool(Tool):
    # default_keymap = 'ctrl+l'
    description = 'Load'


class PreferencesTool(Tool):
    # default_keymap = 'ctrl+e'
    description = 'Preferences'


class QuitTool(Tool):
    # default_keymap = 'q'
    description = 'Quit'


class SaveTool(Tool):
    # default_keymap = 'ctrl+s'
    description = 'Save'


class ViewTool(Tool):
    # default_keymap = 'ctrl+v'
    description = 'View'


class LcapyTkM:

    FIG_WIDTH = 6
    FIG_HEIGHT = 4

    XSIZE = 30
    YSIZE = 20
    SCALE = 0.01

    def __init__(self, filename=None, uimodel_class=None, debug=0):

        self.debug = debug

        # Default Linux backend was TkAgg now QtAgg
        # Default Windows backend Qt4Agg
        import matplotlib.pyplot as p
        if self.debug:
            print(p.get_backend())
        # Need TkAgg if using Tkinter file dialogs
        p.switch_backend('TkAgg')

        super(LcapyTkM, self).__init__()

        from ..uimodelmph import UIModelMPH

        if uimodel_class is None:
            uimodel_class = UIModelMPH

        self.model = uimodel_class(self)

        rcParams['keymap.xscale'].remove('L')
        rcParams['keymap.xscale'].remove('k')
        rcParams['keymap.yscale'].remove('l')
        rcParams['keymap.save'].remove('s')
        rcParams['keymap.save'].remove('ctrl+s')

        rcParams['toolbar'] = 'toolmanager'
        self.fig, self.ax = subplots(1, figsize=(self.FIG_WIDTH,
                                                 self.FIG_HEIGHT))
        self.fig.subplots_adjust(
            left=0.1, top=0.9, bottom=0.1, right=0.9)
        self.ax.axis('equal')

        # Tools to add to the toolbar
        tools = [
            ['Load', LoadTool, self.model.on_load],
            ['Save', SaveTool, self.model.on_save],
            ['Export', ExportTool, self.model.on_export],
            ['View', ViewTool, self.model.on_view],
            ['Inspect', InspectTool, self.model.on_inspect],
            ['Describe', DescribeTool, self.model.on_describe],
            ['Preferences', ExportTool, self.model.on_preferences],
            ['Help', ExportTool, self.model.on_help],
            ['Quit', QuitTool, self.quit]]

        for tool in tools:
            self.fig.canvas.manager.toolmanager.add_tool(tool[0],
                                                         tool[1],
                                                         func=tool[2])
            self.fig.canvas.manager.toolbar.add_tool(tool[0], 'toolgroup')

        self.fig.canvas.mpl_connect('close_event', self.on_close)

        self.draw_grid()

        sketcher = Layer(self.ax)

        self.cursor_sketcher = sketcher
        self.active_sketcher = sketcher
        self.component_sketcher = sketcher
        self.grid_sketcher = sketcher

        self.cid = self.fig.canvas.mpl_connect('button_press_event',
                                               self.on_click_event)

        self.kp_cid = self.fig.canvas.mpl_connect('key_press_event',
                                                  self.on_key_press_event)

        self.fig.canvas.mpl_connect('close_event', self.on_close)

        # Make fullscreen
        self.fig.canvas.manager.full_screen_toggle()

        # self.set_title()

        if filename is not None:
            self.model.load(filename)

    def display(self):

        show()

    def draw_grid(self):

        xticks = arange(self.XSIZE)
        yticks = arange(self.YSIZE)
        self.ax.set_xlim(0, self.XSIZE)
        self.ax.set_ylim(0, self.YSIZE)
        self.ax.set_xticks(xticks)
        self.ax.set_yticks(yticks)
        self.ax.set_xticklabels([])
        self.ax.set_yticklabels([])
        self.ax.grid(color='lightblue')

        self.ax.tick_params(which='both', left=False, bottom=False,
                            top=False, labelbottom=False)

    def clear(self):

        self.component_sketcher.clear()
        self.draw_grid()

    def refresh(self):

        self.fig.canvas.draw()

    def on_key_press_event(self, event):

        # Might be in the textboxes...
        if event.inaxes != self.ax:
            return

        key = event.key
        if self.debug:
            print(key)

        if key in self.model.key_bindings:
            self.model.key_bindings[key]()
        elif key in self.mmodel.key_bindings_with_key:
            self.model.key_bindings_with_key[key](key)

    def on_click_event(self, event):

        if self.debug:
            print('%s click: button=%d, x=%d, y=%d, xdata=%f, ydata=%f' %
                  ('double' if event.dblclick else 'single', event.button,
                   event.x, event.y, event.xdata, event.ydata))

        if event.xdata is None or event.ydata is None:
            return

        if event.dblclick:
            if event.button == 1:
                self.model.on_left_double_click(event.xdata, event.ydata)
            elif event.button == 3:
                self.model.on_right_double_click(event.xdata, event.ydata)
        else:
            if event.button == 1:
                self.model.on_left_click(event.xdata, event.ydata)
            elif event.button == 3:
                self.model.on_right_click(event.xdata, event.ydata)

    def on_close(self, event):

        self.model.on_close()

    def new(self):

        pass

    def quit(self):

        exit()

    def show_expr_dialog(self, expr, title=''):

        from .expr_dialog import ExprDialog

        self.expr_dialog = ExprDialog(expr, self, title)

    def show_inspect_dialog(self, cpt, title=''):

        from .inspect_dialog import InspectDialog

        self.inspect_dialog = InspectDialog(self.model, cpt, title)

    def inspect_properties_dialog(self, cpt, on_changed=None, title=''):

        from .cpt_properties_dialog import CptPropertiesDialog

        self.cpt_properties_dialog = CptPropertiesDialog(cpt,
                                                         on_changed, title)

    def show_node_properties_dialog(self, node, on_changed=None, title=''):

        from .node_properties_dialog import NodePropertiesDialog

        self.node_properties_dialog = NodePropertiesDialog(node,
                                                           on_changed, title)

    def show_plot_properties_dialog(self, expr):

        from .plot_properties_dialog import PlotPropertiesDialog

        self.plot_properties_dialog = PlotPropertiesDialog(expr, self)

    def show_preferences_dialog(self, on_changed=None):

        from .preferences_dialog import PreferencesDialog

        self.preferences_dialog = PreferencesDialog(self, on_changed)

    def show_info_dialog(self, message):

        from tkinter.messagebox import showinfo
        showinfo('', message)

    def show_error_dialog(self, message):

        from tkinter.messagebox import showerror
        showerror('', message)

    def show_message_dialog(self, message, title=''):

        from .message_dialog import MessageDialog

        self.message_dialog = MessageDialog(message, title)

    def open_file_dialog(self, initialdir='.'):

        from tkinter.filedialog import askopenfilename

        filename = askopenfilename(initialdir=initialdir,
                                   title="Select file",
                                   filetypes=(("Lcapy netlist", "*.sch"),))
        return filename

    def save_file_dialog(self, filename):

        from tkinter.filedialog import asksaveasfilename
        from os.path import dirname, splitext, basename

        dirname = dirname(filename)
        basename, ext = splitext(basename(filename))

        options = {}
        options['defaultextension'] = ext
        options['filetypes'] = (("Lcapy netlist", "*.sch"),)
        options['initialdir'] = dirname
        options['initialfile'] = filename
        options['title'] = "Save file"

        return asksaveasfilename(**options)

    def export_file_dialog(self, filename):

        from tkinter.filedialog import asksaveasfilename
        from os.path import dirname, splitext, basename

        dirname = dirname(filename)
        basename, ext = splitext(basename(filename))

        options = {}
        options['defaultextension'] = ext
        options['filetypes'] = (("Embeddable LaTeX", "*.schtex"),
                                ("Standalone LaTeX", "*.tex"),
                                ("PNG image", "*.png"),
                                ("SVG image", "*.svg"),
                                ("PDF", "*.pdf"))
        options['initialdir'] = dirname
        options['initialfile'] = basename + '.pdf'
        options['title'] = "Export file"

        return asksaveasfilename(**options)

    def load(self, filename):
        pass
